

def _itk_warp_affine(image, transform, dsize=None, antialias=False,
                     interpolation='linear', border_mode=None, border_value=0,
                     origin_convention='center', return_info=False):
    """
    ITK backend for warp affine

    References:
        https://examples.itk.org/src/filtering/imagegrid/resampleanimage/documentation
        https://examples.itk.org/src/core/transform/translateimage/documentation
        https://github.com/SimpleITK/TUTORIAL/blob/main/01_spatial_transformations.ipynb
        https://github.com/hinerm/ITK/blob/master/Wrapping/Generators/Python/Tests/ResampleImageFilter.py
        https://simpleitk.readthedocs.io/en/master/fundamentalConcepts.html

    TODO:
        What is the itk package we need to add as a dependency?

    CommandLine:
        xdoctest -m kwimage.im_itk _itk_warp_affine:0 --show
        xdoctest -m kwimage.im_itk _itk_warp_affine:1 --show

    Example:
        >>> # xdoctest: +REQUIRES(module:itk)
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> from kwimage.im_itk import *  # NOQA
        >>> from kwimage.im_itk import _itk_warp_affine
        >>> import numpy as np
        >>> import kwarray
        >>> import kwarray.distributions
        >>> import kwimage
        >>> rng = kwarray.ensure_rng(None)
        >>> image = kwimage.grab_test_image('astro')
        >>> #image = kwimage.ensure_float01(image)
        >>> transform = kwimage.Affine.random(offset=rng.randint(-100, 100))
        >>> # Compare Warp Methods
        >>> #interpolation = 'nearest'
        >>> interpolation = 'linear'
        >>> result1 = kwimage.warp_affine(image, transform, interpolation=interpolation, origin_convention='corner', backend='itk')
        >>> result2 = kwimage.warp_affine(image, transform, interpolation=interpolation, origin_convention='corner', backend='cv2')
        >>> # Test Results
        >>> import kwarray
        >>> print(kwarray.stats_dict(result1))
        >>> print(kwarray.stats_dict(result2))
        >>> result1 = kwimage.ensure_float01(result1)
        >>> result2 = kwimage.ensure_float01(result2)
        >>> difference = (result1 - result2)
        >>> abs_difference = np.abs(difference)
        >>> difference = kwarray.normalize(abs_difference)
        >>> #
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(image, pnum=(1, 4, 1), title='input image', fnum=1, doclf=1)
        >>> kwplot.imshow(result1, pnum=(1, 4, 2), title='itk transform')
        >>> kwplot.imshow(result2, pnum=(1, 4, 3), title='cv2 transform')
        >>> kwplot.imshow(abs_difference, pnum=(1, 4, 4), title='difference')
        >>> #kwplot.imshow((abs_difference > 0).astype(np.float32), pnum=(1, 4, 4), title='difference')
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +SKIP(todo)
        >>> # xdoctest: +REQUIRES(module:itk)
        >>> import kwimage
        >>> import ubelt as ub
        >>> import numpy as np
        >>> #image = kwimage.checkerboard(dsize=(8, 8), num_squares=4, on_value='kitware_blue', off_value='kitware_green')
        >>> image = kwimage.checkerboard(dsize=(32, 32), num_squares=4, on_value='kitware_blue', off_value='kitware_green')
        >>> #image = kwimage.checkerboard(dsize=(4, 4), num_squares=4, on_value='kitware_blue', off_value='kitware_green')
        >>> grid = list(ub.named_product({
        >>>     'border_value': [
        >>>         0,
        >>>         np.nan,
        >>>         #'replicate'
        >>>    ],
        >>>     'interpolation': [
        >>>         'nearest',
        >>>         'linear',
        >>>         #'cubic',
        >>>         #'lanczos',
        >>>         #'area'
        >>>     ],
        >>>     'origin_convention': ['center', 'corner'],
        >>>     'backend': ['itk'],
        >>> }))
        >>> results = []
        >>> results += [('input', image)]
        >>> for kwargs in grid:
        >>>     warped_image = kwimage.warp_affine(image, {'scale': 0.5}, dsize='auto', **kwargs)
        >>>     title = ub.urepr(kwargs, compact=1, nl=1)
        >>>     results.append((title, warped_image))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(results))
        >>> for title, canvas in results:
        >>>     canvas = kwimage.fill_nans_with_checkers(canvas).clip(0, 1)
        >>>     kwplot.imshow(canvas, pnum=pnum_(), origin_convention='corner', title=title, show_ticks=True)
        >>> fig = kwplot.plt.gcf()
        >>> fig.subplots_adjust(hspace=0.37, wspace=0.25)
        >>> fig.set_size_inches([18.11, 11.07])
        >>> kwplot.show_if_requested()

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> # xdoctest: +REQUIRES(module:itk)
        >>> import kwimage
        >>> from kwimage.im_itk import _itk_warp_affine
        >>> from kwimage.transform import Affine
        >>> image = kwimage.grab_test_image('astro')
        >>> transform = Affine.random() @ Affine.translate(300)
        >>> from kwimage.im_cv2 import _cv2_warp_affine
        >>> warped = _itk_warp_affine(image, transform, dsize='auto', interpolation='nearest')
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(warped)
        >>> kwplot.show_if_requested()
    """
    import itk
    import numpy as np
    from kwimage._common import _coerce_warp_dsize_inputs
    if 0:
        [n for n in dir(itk) if 'resample' in n.lower()]
        [n for n in dir(itk) if 'affine' in n.lower()]
        [n for n in dir(itk) if '_from_' in n.lower()]
        [n for n in dir(itk) if 'type' in n.lower()]
        [n for n in dir(itk) if 'RGB' in n.upper()]
        [n for n in dir(itk) if 'UC' in n.upper()]

    # ttype = itk.Image[ itk.UC,  2 ]
    # itk_image = itk.image_from_array(image.T, ttype=ttype)
    itk_image = itk.image_from_array(image.T, is_vector=True)

    # TODO: handle origin convention
    if origin_convention == 'center':
        ...
    elif origin_convention == 'corner':
        itk_image.SetOrigin([0.5, 0.5])
    else:
        raise KeyError(origin_convention)

    # ITK is explcit about image coordinate assumptions.
    output_size = itk.size(itk_image)
    output_spacing = itk.spacing(itk_image)
    output_origin = itk.origin(itk_image)

    h, w = image.shape[0:2]
    input_dsize = (w, h)
    require_warped_info = False
    dsize_inputs = _coerce_warp_dsize_inputs(
        dsize=dsize,
        input_dsize=input_dsize,
        transform=transform,
        require_warped_info=require_warped_info
    )
    # max_dsize = dsize_inputs['max_dsize']
    # new_origin = dsize_inputs['new_origin']
    transform_ = dsize_inputs['transform']
    dsize = dsize_inputs['dsize']
    info = {
        'transform': transform_,
        'dsize': dsize,
        'antialias_info': None,
    }

    numpy_matrix = np.asarray(transform_)
    output_size = dsize

    # Is the above or this perferred?
    # itk_image.UpdateOutputInformation()
    # output_size = itk_image.GetLargestPossibleRegion().GetSize()
    # output_spacing = itk_image.GetSpacing()
    # output_origin = itk_image.GetOrigin()

    # Setup the ITK affine matrix
    itk_transform = itk.AffineTransform.D2.New()
    numpy_matrix = numpy_matrix.astype(np.float64)
    mat_2x2 = numpy_matrix[0:2, 0:2]
    itk_matrix = itk.matrix_from_array(mat_2x2)
    itk_transform.SetMatrix(itk_matrix)
    offset = numpy_matrix[0:2, 2]
    itk_transform.SetTranslation(offset)

    # Setup ITK interpolation
    if interpolation == 'nearest':
        itk_interpolator = itk.NearestNeighborInterpolateImageFunction.New(itk_image)
    elif interpolation == 'linear':
        itk_interpolator = itk.LinearInterpolateImageFunction.New(itk_image)
    else:
        raise NotImplementedError(interpolation)

    if antialias:
        affine_params = transform_.decompose()
        sx, sy = affine_params['scale']
        if sx < 1 or sy < 1:
            # Requires antialias
            raise NotImplementedError(
                'TODO: need to implement antialiasing in ITK backend')
            # AntiAliasFilterType = itk.AntiAliasBinaryImageFilter[ImageType, ImageType]
            # antialiasfilter = AntiAliasFilterType.New()
            # antialiasfilter.SetInput(reader.GetOutput())
            # antialiasfilter.SetMaximumRMSError(args.maximum_RMS_error)
            # antialiasfilter.SetNumberOfIterations(args.number_of_iterations)
            # antialiasfilter.SetNumberOfLayers(args.number_of_layers)

    # Execute transformation
    if 0:
        resampled = itk.resample_image_filter(
            itk_image,
            transform=itk_transform.GetInverseTransform(),
            interpolator=itk_interpolator,
            size=output_size,
            output_spacing=output_spacing,
            output_origin=output_origin,
        )
    else:
        resample_filter = itk.ResampleImageFilter.New(
            itk_image,
            transform=itk_transform.GetInverseTransform(),
            interpolator=itk_interpolator,
            size=output_size,
            output_spacing=output_spacing,
            output_origin=output_origin,
            # default_pixel_value=...,
            # output_direction=...,
            # output_start_index=...,
            # reference_image=...,
            # use_reference_image=...,
        )
        resample_filter.Update()
        resampled = resample_filter.GetOutput()

    # Convert back to numpy
    result = itk.array_from_image(resampled)

    if return_info:
        return result, info
    else:
        return result



"""
Objects for representing and manipulating image transforms, namely image
transformation matrices.
"""
import ubelt as ub
import numpy as np
import kwarray
import skimage.transform
import math
from kwimage import _internal

__all__ = [
    'Transform', 'Matrix', 'Linear', 'Affine', 'Projective'
]


__docstubs__ = """
import affine
"""

try:
    from line_profiler import profile
except Exception:
    profile = ub.identity


class Transform(ub.NiceRepr):
    pass


class Matrix(Transform):
    """
    Base class for matrix-based transform.

    Example:
        >>> from kwimage.transform import *  # NOQA
        >>> ms = {}
        >>> ms['random()'] = Matrix.random()
        >>> ms['eye()'] = Matrix.eye()
        >>> ms['random(3)'] = Matrix.random(3)
        >>> ms['random(4, 4)'] = Matrix.random(4, 4)
        >>> ms['eye(3)'] = Matrix.eye(3)
        >>> ms['explicit'] = Matrix(np.array([[1.618]]))
        >>> for k, m in ms.items():
        >>>     print('----')
        >>>     print(f'{k} = {m}')
        >>>     print(f'{k}.inv() = {m.inv()}')
        >>>     print(f'{k}.T = {m.T}')
        >>>     print(f'{k}.det() = {m.det()}')
    """
    def __init__(self, matrix):
        self.matrix = matrix

    def __nice__(self):
        prefix = '<{}('.format(self.__class__.__name__)
        if isinstance(self.matrix, np.ndarray):
            return np.array2string(self.matrix, separator=', ', prefix=prefix)
        elif self.matrix is None:
            return 'eye'
        else:
            return ub.urepr(self.matrix.tolist(), nl=1)

    def __repr__(self):
        return self.__str__()

    @property
    def shape(self):
        if self.matrix is None:
            # Default shape is hard coded here, can be overrided (e.g. Affine)
            return (2, 2)
        return self.matrix.shape

    def __json__(self):
        if self.matrix is None:
            return {'type': 'matrix', 'matrix': None}
        else:
            return {'type': 'matrix', 'matrix': self.matrix.tolist()}

    @classmethod
    def coerce(cls, data=None, **kwargs):
        """
        Example:
            >>> Matrix.coerce({'type': 'matrix', 'matrix': [[1, 0, 0], [0, 1, 0]]})
            >>> Matrix.coerce(np.eye(3))
            >>> Matrix.coerce(None)
        """
        if data is None and not kwargs:
            return cls(matrix=None)
        if data is None:
            data = kwargs
        if isinstance(data, np.ndarray):
            self = cls(matrix=data)
        elif isinstance(data, cls):
            self = data
        elif data.__class__.__name__ == cls.__name__:
            self = data
        elif isinstance(data, dict):
            keys = set(data.keys())
            if 'matrix' in keys:
                self = cls(matrix=np.array(data['matrix']))
            else:
                raise KeyError(', '.join(list(data.keys())))
        else:
            raise TypeError(type(data))
        return self

    def __array__(self):
        """
        Allow this object to be passed to np.asarray. See [NumpyDispatch]_ for
        details.

        References:
            ..[NumpyDispatch] https://numpy.org/doc/stable/user/basics.dispatch.html
        """
        if self.matrix is None:
            return np.eye(*self.shape)
        return self.matrix

    def __imatmul__(self, other):
        if isinstance(other, np.ndarray):
            other_matrix = other
        else:
            other_matrix = other.matrix
        if self.matrix is None:
            self.matrix = other_matrix
        else:
            self.matrix @= other_matrix

    def __matmul__(self, other):
        """
        Example:
            >>> m = {}
            >>> # Works, and returns a Matrix
            >>> m[len(m)] = x = Matrix.random() @ np.eye(2)
            >>> assert isinstance(x, Matrix)
            >>> m[len(m)] = x = Matrix.random() @ None
            >>> assert isinstance(x, Matrix)
            >>> # Works, and returns an ndarray
            >>> m[len(m)] = x = np.eye(3) @ Matrix.random(3)
            >>> assert isinstance(x, np.ndarray)
            >>> # These do not work
            >>> # m[len(m)] = None @ Matrix.random()
            >>> # m[len(m)] = np.eye(3) @ None
            >>> print('m = {}'.format(ub.urepr(m)))

        Example:
            >>> # Test with rationals
            >>> # xdoctest: +REQUIRES(module:sympy)
            >>> import kwimage
            >>> a = kwimage.Matrix.random((3, 3)).rationalize()
            >>> b = kwimage.Matrix.random((3, 3)).rationalize()
            >>> c = kwimage.Matrix.random((3, 3))
            >>> assert not c.is_rational()
            >>> assert (a @ c).is_rational()
            >>> assert (c @ a).is_rational()

        Ignore:
            %timeit c.inv()
            %timeit a.inv()
        """
        if other is None:
            return self
        if self.matrix is None:
            return self.__class__.coerce(other)
        if isinstance(other, np.ndarray):
            return self.__class__(self.matrix @ other)
        elif other.matrix is None:
            return self
        elif isinstance(other, self.__class__):
            # Prefer using the type of the left-hand-side, but try
            # not to break group rules.
            return self.__class__(self.matrix @ other.matrix)
        elif isinstance(self, other.__class__):
            return other.__class__(self.matrix @ other.matrix)
        else:
            raise TypeError('{} @ {}'.format(type(self), type(other)))

    def is_rational(self):
        """
        TODO: rename to "is_symbolic"
        """
        if sympy is None:
            return False
        return isinstance(self.matrix, sympy.Matrix) and isinstance(self.matrix[0, 0], (sympy.Rational, sympy.core.symbol.Symbol, sympy.core.basic.Basic))

    def inv(self):
        """
        Returns the inverse of this matrix

        Returns:
            Matrix

        Example:
            >>> # Test with rationals
            >>> # xdoctest: +REQUIRES(module:sympy)
            >>> import kwimage
            >>> self = kwimage.Matrix.random((3, 3)).rationalize()
            >>> inv = self.inv()
            >>> eye = self @ inv
            >>> eye.isclose_identity(0, 0)
        """
        if self.matrix is None:
            return self.__class__(None)
        else:
            try:
                inv_mat = np.linalg.inv(self.matrix)
            except (np.linalg.LinAlgError, TypeError):
                # using TypeError instead of np.core._exceptions.UFuncTypeError
                if self.is_rational():
                    # inv_mat = mp.inverse(self.matrix)
                    # handle object arrays (rationals)
                    inv_mat = self.matrix.inv()
                else:
                    raise
            return self.__class__(inv_mat)

    @property
    def T(self):
        """
        Transpose the underlying matrix
        """
        if self.matrix is None:
            return self
        else:
            return self.__class__(self.matrix.T)

    def det(self):
        """
        Compute the determinant of the underlying matrix

        Returns:
            float
        """
        if self.matrix is None:
            return 1.
        else:
            try:
                det = np.linalg.det(self.matrix)
            except np.core._exceptions.UFuncTypeError:
                # handle object arrays (rationals)
                det = self.matrix.det()
            return det

    @classmethod
    def eye(cls, shape=None, rng=None):
        """
        Construct an identity
        """
        self = cls(None)
        if isinstance(shape, int):
            shape = (shape, shape)
        if shape is None:
            shape = self.shape
        self.matrix = np.eye(*shape)
        return self

    @classmethod
    def random(cls, shape=None, rng=None):
        import kwarray
        rng = kwarray.ensure_rng(rng)
        self = cls(None)
        if isinstance(shape, int):
            shape = (shape, shape)
        if shape is None:
            shape = self.shape
        self.matrix = rng.rand(*shape)
        return self

    def __getitem__(self, index):
        if self.matrix is None:
            return np.asarray(self)[index]
        return self.matrix[index]

    def rationalize(self):
        """
        Convert the underlying matrix to a rational type to avoid floating
        point errors. This does decrease efficiency.

        TODO:
            - [ ] mpmath for arbitrary precision? It doesn't seem to do
            inverses correct whereas sympy does.

        Ignore:
            from sympy import Rational
            from fractions import Fraction
            float_mat = np.random.rand(3, 3)
            float_num = float_mat[0, 0]
            frac_num = Fraction(float_num)
            rat_num = Rational(float_num)

            rat_mat_v1 = float_mat.astype(Rational, subok=False)
            frac_mat_v1 = float_mat.astype(Fraction, subok=False)
            rat_num_v1 = rat_mat_v1[0, 0]
            frac_num_v1 = frac_mat_v1[0, 0]
            print(f'{type(float_num)=}')
            print(f'{type(rat_num)=}')
            print(f'{type(frac_num)=}')
            print(f'{type(rat_num_v1)=}')
            print(f'{type(frac_num_v1)=}')

            flat_rat = list(map(Rational, float_mat.ravel().tolist()))
            from sympy import Matrix
            rat_mat_v2 = Matrix(flat_rat).reshape(*float_mat.shape)

        Example:
            >>> # xdoctest: +REQUIRES(module:sympy)
            >>> import kwimage
            >>> self = kwimage.Matrix.random((3, 3))
            >>> mat = self.rationalize()
            >>> mat2 = kwimage.Matrix.random((3, 3))
            >>> mat3 = mat @ mat2
            >>> #assert 'sympy' in mat3.matrix.__class__.__module__
            >>> mat3 = mat2 @ mat
            >>> #assert 'sympy' in mat3.matrix.__class__.__module__
            >>> assert not mat.isclose_identity()
            >>> assert (mat @ mat.inv()).isclose_identity(rtol=0, atol=0)
        """
        if self.matrix is None:
            new_mat = self.matrix
        else:
            new_mat = _RationalNDArray.from_numpy(self.matrix)
        new = self.__class__(new_mat)
        return new

    def astype(self, dtype):
        """
        Convert the underlying matrix to a rational type to avoid floating
        point errors. This does decrease efficiency.

        Args:
            dtype (type):
        """
        if self.matrix is None:
            new_mat = self.matrix
        else:
            new_mat = self.matrix.astype(dtype)
        new = self.__class__(new_mat)
        return new

    def isclose_identity(self, rtol=1e-05, atol=1e-08):
        """
        Returns true if the matrix is nearly the identity.
        """
        if self.matrix is None:
            return True
        else:
            eye = np.eye(*self.matrix.shape)
            try:
                return np.allclose(self.matrix, eye, rtol=rtol, atol=atol)
            except TypeError:
                if self.is_rational():
                    residual = np.array(self.matrix - eye).astype(float)
                    return np.allclose(residual, 0, rtol=rtol, atol=atol)
                else:
                    raise


class Linear(Matrix):
    pass


class Projective(Linear):
    """
    A thin wraper around a 3x3 matrix that represent a projective transform

    Implements methods for:
        * creating random projective transforms
        * decomposing the matrix
        * finding a best-fit transform between corresponding points
        * TODO: - [ ] fully rational transform

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import math
        >>> image = kwimage.grab_test_image()
        >>> theta = 0.123 * math.tau
        >>> components = {
        >>>     'rotate': kwimage.Projective.projective(theta=theta),
        >>>     'scale': kwimage.Projective.projective(scale=0.5),
        >>>     'shear': kwimage.Projective.projective(shearx=0.2),
        >>>     'translation': kwimage.Projective.projective(offset=(100, 200)),
        >>>     'rotate+translate': kwimage.Projective.projective(theta=0.123 * math.tau, about=(256, 256)),
        >>>     'perspective': kwimage.Projective.projective(uv=(0.0003, 0.0007)),
        >>>     'random-composed': kwimage.Projective.random(scale=(0.5, 1.5), translate=(-20, 20), theta=(-theta, theta), shearx=(0, .4), rng=900558176210808600),
        >>> }
        >>> warp_stack = []
        >>> for key, mat in components.items():
        ...     warp = kwimage.warp_projective(image, mat)
        ...     warp = kwimage.draw_text_on_image(
        ...        warp,
        ...        ub.urepr(mat.matrix, nl=1, nobr=1, precision=4, si=1, sv=1, with_dtype=0),
        ...        org=(1, 1),
        ...        valign='top', halign='left',
        ...        fontScale=0.8, color='kw_green',
        ...        border={'thickness': 3},
        ...        )
        ...     warp = kwimage.draw_header_text(warp, key, color='kw_blue')
        ...     warp_stack.append(warp)
        >>> warp_canvas = kwimage.stack_images_grid(warp_stack, chunksize=4, pad=10, bg_value='kitware_gray')
        >>> # xdoctest: +REQUIRES(module:sympy)
        >>> import sympy
        >>> # Shows the symbolic construction of the code
        >>> # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
        >>> from sympy.abc import theta
        >>> params = x0, y0, sx, sy, theta, shearx, tx, ty, u, v = sympy.symbols(
        >>>     'x0, y0, sx, sy, theta, ex, tx, ty, u, v')
        >>> # move the center to 0, 0
        >>> tr1_ = sympy.Matrix([[1, 0,  -x0],
        >>>                      [0, 1,  -y0],
        >>>                      [0, 0,    1]])
        >>> P = sympy.Matrix([  # projective part
        >>>     [ 1,  0,  0],
        >>>     [ 0,  1,  0],
        >>>     [ u,  v,  1]])
        >>> # Define core components of the affine transform
        >>> S = sympy.Matrix([  # scale
        >>>     [sx,  0, 0],
        >>>     [ 0, sy, 0],
        >>>     [ 0,  0, 1]])
        >>> E = sympy.Matrix([  # x-shear
        >>>     [1,  shearx, 0],
        >>>     [0,  1, 0],
        >>>     [0,  0, 1]])
        >>> R = sympy.Matrix([  # rotation
        >>>     [sympy.cos(theta), -sympy.sin(theta), 0],
        >>>     [sympy.sin(theta),  sympy.cos(theta), 0],
        >>>     [               0,                 0, 1]])
        >>> T = sympy.Matrix([  # translation
        >>>     [ 1,  0, tx],
        >>>     [ 0,  1, ty],
        >>>     [ 0,  0,  1]])
        >>> # move 0, 0 back to the specified origin
        >>> tr2_ = sympy.Matrix([[1, 0,  x0],
        >>>                      [0, 1,  y0],
        >>>                      [0, 0,   1]])
        >>> # combine transformations
        >>> homog_ = sympy.MatMul(tr2_, T, R, E, S, P, tr1_)
        >>> #with sympy.evaluate(False):
        >>> #    homog_ = sympy.MatMul(tr2_, T, R, E, S, P, tr1_)
        >>> #    sympy.pprint(homog_)
        >>> homog = homog_.doit()
        >>> #sympy.pprint(homog)
        >>> print('homog = {}'.format(ub.urepr(homog.tolist(), nl=1)))
        >>> # This could be prettier
        >>> texts = {
        >>>     'Translation': sympy.pretty(T, use_unicode=0),
        >>>     'Rotation': sympy.pretty(R, use_unicode=0),
        >>>     'shEar-X': sympy.pretty(E, use_unicode=0),
        >>>     'Scale': sympy.pretty(S, use_unicode=0),
        >>>     'Perspective': sympy.pretty(P, use_unicode=0),
        >>> }
        >>> print(ub.urepr(texts, nl=2, sv=1))
        >>> equation_stack = []
        >>> for text, m in texts.items():
        >>>     render_canvas = kwimage.draw_text_on_image(None, m, color='kw_green', fontScale=1.0)
        >>>     render_canvas = kwimage.draw_header_text(render_canvas, text, color='kw_blue')
        >>>     render_canvas = kwimage.imresize(render_canvas, scale=1.3)
        >>>     equation_stack.append(render_canvas)
        >>> equation_canvas = kwimage.stack_images(equation_stack, pad=10, axis=1, bg_value='kitware_gray')
        >>> render_canvas = kwimage.draw_text_on_image(None, sympy.pretty(homog, use_unicode=0), color='kw_green', fontScale=1.0)
        >>> render_canvas = kwimage.draw_header_text(render_canvas, 'Full Equation With Pre-Shift', color='kw_blue')
        >>> # xdoctest: -REQUIRES(module:sympy)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> canvas = kwimage.stack_images([warp_canvas, equation_canvas, render_canvas], pad=20, axis=0, bg_value='kitware_gray', resize='larger')
        >>> canvas = kwimage.draw_header_text(canvas, 'Projective matrixes can represent', color='kw_blue')
        >>> kwplot.imshow(canvas)
        >>> fig = plt.gcf()
        >>> fig.set_size_inches(13, 13)
        >>> kwplot.show_if_requested()
    """
    # References:
    #     .. [AffineDecompColab] https://colab.research.google.com/drive/1ImBB-N6P9zlNMCBH9evHD6tjk0dzvy1_

    @classmethod
    def fit(cls, pts1, pts2):
        """
        Fit an projective transformation between a set of corresponding points.

        See [HomogEst]_ [SzeleskiBook]_ and [RansacDummies]_ for references on
        the subject.

        Args:
            pts1 (ndarray): An Nx2 array of points in "space 1".
            pts2 (ndarray): A corresponding Nx2 array of points in "space 2"

        Returns:
            Projective : a transform that warps from "space1" to "space2".

        Note:
            A projective matrix has 8 degrees of freedom, so at least 8 point
            pairs are needed.

        References:
            .. [HomogEst] http://dip.sun.ac.za/~stefan/TW793/attach/notes/homography_estimation.pdf
            .. [SzeleskiBook] http://szeliski.org/Book/drafts/SzeliskiBook_20100903_draft.pdf Page 317
            .. [RansacDummies] http://vision.ece.ucsb.edu/~zuliani/Research/RANSAC/docs/RANSAC4Dummies.pdf page 53

        Example:
            >>> # Create a set of points, warp them, then recover the warp
            >>> import kwimage
            >>> points = kwimage.Points.random(9).scale(64)
            >>> A1 = kwimage.Affine.affine(scale=0.9, theta=-3.2, offset=(2, 3), about=(32, 32), skew=2.3)
            >>> A2 = kwimage.Affine.affine(scale=0.8, theta=0.8, offset=(2, 0), about=(32, 32))
            >>> A12_real = A2 @ A1.inv()
            >>> points1 = points.warp(A1)
            >>> points2 = points.warp(A2)
            >>> # Make the correspondence non-affine
            >>> points2.data['xy'].data[0, 0] += 3.5
            >>> points2.data['xy'].data[3, 1] += 8.5
            >>> # Recover the warp
            >>> pts1, pts2 = points1.xy, points2.xy
            >>> A_recovered = kwimage.Projective.fit(pts1, pts2)
            >>> #assert np.all(np.isclose(A_recovered.matrix, A12_real.matrix))
            >>> # xdoctest: +REQUIRES(--show)
            >>> import cv2
            >>> import kwplot
            >>> kwplot.autompl()
            >>> base1 = np.zeros((96, 96, 3))
            >>> base1[32:-32, 5:-5] = 0.5
            >>> base2 = np.zeros((96, 96, 3))
            >>> img1 = points1.draw_on(base1, radius=3, color='blue')
            >>> img2 = points2.draw_on(base2, radius=3, color='green')
            >>> img1_warp = kwimage.warp_projective(img1, A_recovered.matrix, dsize=img1.shape[0:2][::-1])
            >>> canvas = kwimage.stack_images([img1, img2, img1_warp], pad=10, axis=1, bg_value=(1., 1., 1.))
            >>> kwplot.imshow(canvas)
        """
        if 0:
            import cv2
            inlier_method = 'all'
            inlier_method_lut = {
                'all': 0,
                'lmeds': cv2.LMEDS,
                'ransac': cv2.RANSAC,
                'prosac': cv2.RHO,
            }
            cv2_method = inlier_method_lut[inlier_method]
            # This probably does the point normaliztion internally,
            # but I'm not sure
            H, mask = cv2.findHomography(pts1, pts2, method=cv2_method)
            return Projective(H)
        else:
            def whiten_xy_points(xy_m):
                """
                whitens points to mean=0, stddev=1 and returns transformation
                """
                mu_xy  = xy_m.mean(axis=1)  # center of mass
                std_xy = xy_m.std(axis=1)
                std_xy[std_xy == 0] = 1  # prevent divide by zero
                tx, ty = -mu_xy / std_xy
                sx, sy = 1 / std_xy
                T = np.array([(sx, 0, tx),
                              (0, sy, ty),
                              (0,  0,  1)])
                xy_norm = ((xy_m.T - mu_xy) / std_xy).T
                return xy_norm, T
            # Hartley Precondition (to reduce sensitivity to noise)
            xy1_mn, T1 = whiten_xy_points(pts1.T)
            xy2_mn, T2 = whiten_xy_points(pts2.T)
            # xy1_mn = pts1.T
            # xy2_mn = pts2.T
            x1_mn = xy1_mn[0]
            y1_mn = xy1_mn[1]
            x2_mn = xy2_mn[0]
            y2_mn = xy2_mn[1]
            num_pts = x1_mn.shape[0]
            # Concatenate all 2x9 matrices into an Mx9 matrix
            Mx9 = np.empty((2 * num_pts, 9), dtype=float)
            for ix in range(num_pts):
                u2        = x2_mn[ix]
                v2        = y2_mn[ix]
                x1        = x1_mn[ix]
                y1        = y1_mn[ix]
                (d, e, f) = (     -x1,      -y1,  -1)
                (g, h, i) = ( v2 * x1,  v2 * y1,  v2)
                (j, k, l) = (      x1,       y1,   1)
                (p, q, r) = (-u2 * x1, -u2 * y1, -u2)
                Mx9[ix * 2]     = (0, 0, 0, d, e, f, g, h, i)
                Mx9[ix * 2 + 1] = (j, k, l, 0, 0, 0, p, q, r)
            M = (Mx9.T @ Mx9)
            # M = Mx9
            try:
                # https://math.stackexchange.com/questions/772039/how-does-the-svd-solve-the-least-squares-problem/2173715#2173715
                # http://twistedoakstudios.com/blog/Post7254_visualizing-the-eigenvectors-of-a-rotation
                USVt = np.linalg.svd(M, full_matrices=True, compute_uv=True)
            except MemoryError:
                import scipy.sparse as sps
                import scipy.sparse.linalg as spsl
                M_sparse = sps.lil_matrix(M)
                USVt = spsl.svds(M_sparse)
            except np.linalg.LinAlgError:
                raise
            except Exception:
                raise
            # U is the co-domain unitary matrix
            # V is the domain unitary matrix
            # s contains the singular values
            U, s, Vt = USVt
            # The column of V (row of Vt) corresponding to the lowest singular
            # value is the solution to the least squares problem
            H_prime = Vt[8].reshape(3, 3)

            # Then compute ax = b  [aka: x = npl.solve(a, b)]
            M = np.linalg.inv(T2) @ H_prime @ T1  # Unnormalize
            # homographies that only differ by a scale factor are equivalent
            M /= M[2, 2]
            return Projective(M)

    @classmethod
    def projective(cls, scale=None, offset=None, shearx=None, theta=None,
                   uv=None, about=None):
        """
        Reconstruct from parameters

        Sympy:
            >>> # xdoctest: +SKIP
            >>> import sympy
            >>> # Shows the symbolic construction of the code
            >>> # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
            >>> from sympy.abc import theta
            >>> params = x0, y0, sx, sy, theta, shearx, tx, ty, u, v = sympy.symbols(
            >>>     'x0, y0, sx, sy, theta, ex, tx, ty, u, v')
            >>> # move the center to 0, 0
            >>> tr1_ = sympy.Matrix([[1, 0,  -x0],
            >>>                      [0, 1,  -y0],
            >>>                      [0, 0,    1]])
            >>> P = sympy.Matrix([  # projective part
            >>>     [ 1,  0,  0],
            >>>     [ 0,  1,  0],
            >>>     [ u,  v,  1]])
            >>> # Define core components of the affine transform
            >>> S = sympy.Matrix([  # scale
            >>>     [sx,  0, 0],
            >>>     [ 0, sy, 0],
            >>>     [ 0,  0, 1]])
            >>> E = sympy.Matrix([  # x-shear
            >>>     [1,  shearx, 0],
            >>>     [0,  1, 0],
            >>>     [0,  0, 1]])
            >>> R = sympy.Matrix([  # rotation
            >>>     [sympy.cos(theta), -sympy.sin(theta), 0],
            >>>     [sympy.sin(theta),  sympy.cos(theta), 0],
            >>>     [               0,                 0, 1]])
            >>> T = sympy.Matrix([  # translation
            >>>     [ 1,  0, tx],
            >>>     [ 0,  1, ty],
            >>>     [ 0,  0,  1]])
            >>> # move 0, 0 back to the specified origin
            >>> tr2_ = sympy.Matrix([[1, 0,  x0],
            >>>                      [0, 1,  y0],
            >>>                      [0, 0,   1]])
            >>> # combine transformations
            >>> with sympy.evaluate(False):
            >>>     homog_ = sympy.MatMul(tr2_, T, R, E, S, P, tr1_)
            >>>     sympy.pprint(homog_)
            >>> homog = homog_.doit()
            >>> sympy.pprint(homog)
            >>> print('homog = {}'.format(ub.urepr(homog.tolist(), nl=1)))

        Ignore:
            M = kwimage.Projective.projective(uv=(0, 0.04), about=128)
            img1 = kwimage.ensure_float01(kwimage.grab_test_image('astro', dsize=(228, 228)))
            points = kwimage.Points(xy=kwimage.Coords(np.array([
                (0, 0),
                (1, 1),
                (128, 1),
                (144, 0),
                (288, 288),
                (97, 77),
                (0, 288),
                (128, 128),
            ])))
            img1_warp = kwimage.warp_projective(img1, M.matrix, interpolation='linear')
            warped = points.warp(M.matrix)

            pic_copy = points.draw_on(img1.copy(), radius=10, color='kitware_green')
            warp_copy = warped.draw_on(img1_warp.copy(), radius=10, color='kitware_green')
            stacked, stack_tfs = kwimage.stack_images([pic_copy, warp_copy], return_info=True, axis=1)
            stacked = kwimage.draw_line_segments_on_image(stacked, points.warp(stack_tfs[0]).xy, warped.warp(stack_tfs[1]).xy)

            import kwplot
            kwplot.autompl()
            kwplot.imshow(stacked)
            # M.matrix.dot(np.array([[0, 0, 1]]).T)
        """
        import kwimage
        about_ = 0 if about is None else about
        if uv is None:
            uv = 0, 0
        x0, y0 = _ensure_iterable2(about_)
        # About needs to be wrt to this because the projective and affine parts
        # will be inside it.
        tr1_ = np.array([[1, 0,  -x0],
                         [0, 1,  -y0],
                         [0, 0,    1]])
        tr2_ = np.array([[1, 0,  x0],
                         [0, 1,  y0],
                         [0, 0,   1]])
        # TODO: add sympy optimization
        aff_part = kwimage.Affine.affine(
            scale=scale, offset=offset, shearx=shearx, theta=theta)
        u, v = uv
        proj_part = np.array([
            [ 1,  0,  0],
            [ 0,  1,  0],
            [ u,  v,  1],
        ])
        self = kwimage.Projective(tr2_ @ aff_part.matrix @ proj_part @ tr1_)
        return self

    @classmethod
    def coerce(cls, data=None, **kwargs):
        """
        Attempt to coerce the data into an Projective object

        Args:
            data : some data we attempt to coerce to an Projective matrix
            **kwargs : some data we attempt to coerce to an Projective matrix,
                mutually exclusive with `data`.

        Returns:
            Projective

        Example:
            >>> import kwimage
            >>> kwimage.Projective.coerce({'type': 'affine', 'matrix': [[1, 0, 0], [0, 1, 0]]})
            >>> kwimage.Projective.coerce({'type': 'affine', 'scale': 2})
            >>> kwimage.Projective.coerce({'type': 'projective', 'scale': 2})
            >>> kwimage.Projective.coerce({'scale': 2})
            >>> kwimage.Projective.coerce({'offset': 3})
            >>> kwimage.Projective.coerce(np.eye(3))
            >>> kwimage.Projective.coerce(None)
            >>> import skimage
            >>> kwimage.Projective.coerce(skimage.transform.AffineTransform(scale=30))
            >>> kwimage.Projective.coerce(skimage.transform.ProjectiveTransform(matrix=None))
        """
        if data is None and not kwargs:
            return cls(matrix=None)
        if data is None:
            data = kwargs
        if isinstance(data, np.ndarray):
            self = cls(matrix=data)
        elif isinstance(data, cls):
            self = data
        elif isinstance(data, (skimage.transform.AffineTransform, skimage.transform.ProjectiveTransform)):
            self = cls(matrix=data.params)
        elif data.__class__.__name__ == cls.__name__:
            self = data
        elif isinstance(data, dict):
            keys = set(data.keys())
            if 'matrix' in keys:
                matrix = np.array(data['matrix'])
                if matrix.shape[0] == 2:
                    matrix = np.vstack([matrix, [[0, 0, 1.]]])
                self = cls(matrix=matrix)
            else:
                known_params = {'uv', 'scale', 'offset', 'theta', 'type', 'shearx', 'shear', 'about'}
                params = {key: data[key] for key in known_params if key in data}
                if len(keys - known_params) == 0:
                    type_ = params.pop('type', None)  # NOQA
                    # if len(keys) == 1:
                    #     # Special cases for speed
                    #     if keys == {'scale'}:
                    #         self = cls.scale(**params)
                    #     if keys == {'translate'}:
                    #         self = cls.scale(**params)
                    #     else:
                    #         self = cls.projective(**params)
                    # else:
                    self = cls.projective(**params)
                else:
                    raise KeyError(', '.join(list(data.keys())))
        else:
            raise TypeError(type(data))
        return self

    def is_affine(self):
        """
        If the bottom row is [[0, 0, 1]], then this can be safely turned into
        an affine matrix.

        Returns:
            bool

        Example:
            >>> import kwimage
            >>> bool(kwimage.Projective.coerce(scale=2, uv=[1, 1]).is_affine())
            False
            >>> bool(kwimage.Projective.coerce(scale=2, uv=[0, 0]).is_affine())
            True
        """
        if self.matrix is None:
            return True
        else:
            return np.all(self.matrix[2] == [0, 0, 1])

    def to_skimage(self):
        """
        Returns:
            skimage.transform.AffineTransform

        Example:
            >>> import kwimage
            >>> self = kwimage.Projective.random()
            >>> tf = self.to_skimage()
            >>> # Transform points with kwimage and scikit-image
            >>> kw_poly = kwimage.Polygon.random()
            >>> kw_warp_xy = kw_poly.warp(self.matrix).exterior.data
            >>> sk_warp_xy = tf(kw_poly.exterior.data)
            >>> assert np.allclose(sk_warp_xy, sk_warp_xy)
        """
        return skimage.transform.ProjectiveTransform(matrix=np.asarray(self))

    @classmethod
    def random(cls, shape=None, rng=None, **kw):
        """
        Example/
            >>> import kwimage
            >>> self = kwimage.Projective.random()
            >>> print(f'self={self}')
            >>> params = self.decompose()
            >>> aff_part = kwimage.Affine.affine(**ub.dict_diff(params, ['uv']))
            >>> proj_part = kwimage.Projective.coerce(uv=params['uv'])
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import cv2
            >>> import kwplot
            >>> dsize = (256, 256)
            >>> kwplot.autompl()
            >>> img1 = kwimage.grab_test_image(dsize=dsize)
            >>> img1_affonly = kwimage.warp_projective(img1, aff_part.matrix, dsize=img1.shape[0:2][::-1])
            >>> img1_projonly = kwimage.warp_projective(img1, proj_part.matrix, dsize=img1.shape[0:2][::-1])
            >>> ###
            >>> img2 = kwimage.ensure_uint255(kwimage.atleast_3channels(kwimage.checkerboard(dsize=dsize)))
            >>> img1_fullwarp = kwimage.warp_projective(img1, self.matrix, dsize=img1.shape[0:2][::-1])
            >>> img2_affonly = kwimage.warp_projective(img2, aff_part.matrix, dsize=img2.shape[0:2][::-1])
            >>> img2_projonly = kwimage.warp_projective(img2, proj_part.matrix, dsize=img2.shape[0:2][::-1])
            >>> img2_fullwarp = kwimage.warp_projective(img2, self.matrix, dsize=img2.shape[0:2][::-1])
            >>> canvas1 = kwimage.stack_images([img1, img1_projonly, img1_affonly, img1_fullwarp], pad=10, axis=1, bg_value=(0.5, 0.9, 0.1))
            >>> canvas2 = kwimage.stack_images([img2, img2_projonly, img2_affonly, img2_fullwarp], pad=10, axis=1, bg_value=(0.5, 0.9, 0.1))
            >>> canvas = kwimage.stack_images([canvas1, canvas2], axis=0)
            >>> kwplot.imshow(canvas)
        """
        import kwimage
        rng = kwarray.ensure_rng(rng)
        aff_part = kwimage.Affine.random(shape, rng=rng, **kw)
        # Random projective part
        u = 1 / rng.randint(1, 10000)
        v = 1 / rng.randint(1, 10000)
        proj_part = np.array([
            [ 1,  0,  0],
            [ 0,  1,  0],
            [ u,  v,  1],
        ])
        self = Projective(aff_part.matrix @ proj_part)
        return self

    def decompose(self):
        r"""
        Based on the analysis done in [ME1319680]_.

        Returns:
            Dict:

        References:
            .. [ME1319680] https://math.stackexchange.com/questions/1319680

        Example:
            >>> # Create a set of points, warp them, then recover the warp
            >>> import kwimage
            >>> points = kwimage.Points.random(9).scale(64)
            >>> A1 = kwimage.Affine.affine(scale=0.9, theta=-3.2, offset=(2, 3), about=(32, 32), skew=2.3)
            >>> A2 = kwimage.Affine.affine(scale=0.8, theta=0.8, offset=(2, 0), about=(32, 32))
            >>> A12_real = A2 @ A1.inv()
            >>> points1 = points.warp(A1)
            >>> points2 = points.warp(A2)
            >>> # Make the correspondence non-affine
            >>> points2.data['xy'].data[0, 0] += 3.5
            >>> points2.data['xy'].data[3, 1] += 8.5
            >>> # Recover the warp
            >>> pts1, pts2 = points1.xy, points2.xy
            >>> self = kwimage.Projective.random()
            >>> self.decompose()

        Example:
            >>> # xdoctest: +REQUIRES(module:sympy)
            >>> from kwimage.transform import *  # NOQA
            >>> import kwimage
            >>> from kwimage.transform import _RationalNDArray
            >>> self = kwimage.Projective.random().rationalize()
            >>> rat_decomp = self.decompose()
            >>> print('rat_decomp = {}'.format(ub.urepr(rat_decomp, nl=1)))
            >>> ####
            >>> import sympy
            >>> cells = sympy.symbols('h1, h2, h3, h4, h5, h6, h7, h8, h9')
            >>> matrix = _RationalNDArray(cells).reshape(3, 3)
            >>> # Symbolic decomposition. Neat.
            >>> self = kwimage.Projective(matrix)
            >>> self.decompose()

        Ignore:
            >>> from sympy.abc import theta
            >>> import sympy
            >>> h1, h2, h3, h4, h5, h6, h7, h8, h9 = sympy.symbols(
            >>>     'h1, h2, h3, h4, h5, h6, h7, h8, h9')
            >>> H = sympy.Matrix([[h1, h2, h3], [h4, h5, h6], [h7, h8, 1]])
            >>> a1 = h1 - h3 * h7
            >>> a2 = h2 - h3 * h8
            >>> a3 = h3
            >>> a4 = h4 - h6 * h7
            >>> a5 = h5 - h6 * h8
            >>> a6 = h6
            >>> A = sympy.Matrix([[a1, a2, a3], [a4, a5, a6], [0, 0, 1]])
            >>> P = sympy.Matrix([[1, 0, 0], [0, 1, 0], [h7, h8, 1]])
            >>> assert np.all(np.ravel((A @ P - H).tolist()) == 0)

            # TODO: Can we get a more concise ane nice sympy decomposition /
            # recombination
            sympy.printing.pretty_print(sympy.Eq(H, B @ Q))
            bs = b1, b2, b3, b4, b5, b6, b7, b8, b9 = sympy.symbols(
            'b1, b2, b3, b4, b5, b6, b7, b8, b9')
            uvs = u, v = sympy.symbols('u, v')
            B = sympy.Matrix([[b1, b2, b3], [b4, b5, b6], [0, 0, 1]])
            Q = sympy.Matrix([[1, 0, 0], [0, 1, 0], [u, v, 1]])
            # Q = sympy.Matrix([[1, 0, 0], [0, 1, 0], [u, v, 1 / (b3*u + b6*v + 1)]])
            H2_unnorm = Q @ B
            H2_norm = H2_unnorm / H2_unnorm.tolist()[-1][-1]
            expr = sympy.Eq(H, Q @ B)
            sympy.solve(expr, bs + uvs)
            A @ P = H
            A @ P = H =
            A @ P @ A.inv() @ A
            A @ P @ A.inv()

            kwimage.Projective.projective(

        Ignore:
            import kwimage
            import kwplot
            import sympy
            plt = kwplot.autoplt()
            from kwplot.cli import gifify
            import cv2
            check = kwimage.atleast_3channels(kwimage.checkerboard(dsize=(288, 288)))
            pic = kwimage.grab_test_image('astro', dsize=check.shape[0:2][::-1])
            img1 = np.maximum((1 - check) * kwimage.ensure_float01(pic), check)
            ims = []
            # v_coords = np.log(np.logspace(1, 2)) / np.log(10) - 1
            v_coords = np.linspace(0, 1.0, 64) ** 6
            for v in ub.ProgIter(v_coords):
                u = 0
                v = v
                I = kwimage.Affine.eye()
                T = kwimage.Affine.translate(-128)
                # T = kwimage.Affine.translate(0)
                H = kwimage.Projective(np.array([[1, 0, 0], [0, 1, 0], [u, v, 1]]).astype(np.float32))
                H2 = (T.inv() @ H @ T)
                C2 = (I @ I)
                img1_warp = kwimage.warp_projective(img1, H2.matrix, dsize=img1.shape[0:2][::-1]).clip(0, 1)
                img1_pre = kwimage.warp_projective(img1, C2.matrix, dsize=img1.shape[0:2][::-1]).clip(0, 1)
                canvas = kwimage.stack_images([img1_pre, img1_warp], pad=10, axis=1, bg_value=(0., 1., 0.))
                canvas = kwimage.draw_text_on_image(
                    canvas,
                    'u={},{}v={}'.format(round(u, 8), chr(10) * 2, round(v, 8)),
                    # org=tuple(img1.shape[0:2][::-1]),
                    # valign='bottom', halign='right',
                    org=(1, 1),
                    valign='top', halign='left',
                    fontScale=0.8,
                    border=True,
                    )
                ims.append(canvas)
            # Hinge animation
            images = ims
            dpath = ub.Path.appdir('kwcoco/demo').ensuredir()
            output_fpath = dpath / 'hinge-v-t.gif'
            gifify.ffmpeg_animate_images(ims, output_fpath, in_framerate=2)
        """
        import numpy as np
        h1, h2, h3, h4, h5, h6, h7, h8, h9 = self.matrix.ravel()
        # assert h9 == 1

        a1 = h1 - h3 * h7
        a2 = h2 - h3 * h8
        a3 = h3
        a4 = h4 - h6 * h7
        a5 = h5 - h6 * h8
        a6 = h6

        mcls = _RationalNDArray if self.is_rational() else np.array

        affine_part = Affine(mcls([
            [a1, a2, a3],
            [a4, a5, a6],
            [0,   0,  1],
        ]))
        decomp = affine_part.decompose()
        # The line u * x + v * y = 0 is fixed to iteself.
        # I.e. y = -u/v * x + 0

        # The line u * x + v * y = -1 is mapped to the point at infinity
        # I.e. y = -u/v * x - 1/v
        u, v = h7, h8
        # This transform has to happen first when we re-compose
        # TODO: I would love to find a more intuitive name or representation
        # for this. Can I do something to call this a "hinge"? Is there a
        # representation where I can look at this and get a sense of where the
        # "hinge" is?
        decomp['uv'] = (u, v)
        return decomp


class Affine(Projective):
    """
    A thin wraper around a 3x3 matrix that represents an affine transform

    Implements methods for:
        * creating random affine transforms
        * decomposing the matrix
        * finding a best-fit transform between corresponding points
        * TODO: - [ ] fully rational transform

    Example:
        >>> # xdoctest: +REQUIRES(module:cv2)
        >>> import kwimage
        >>> import math
        >>> image = kwimage.grab_test_image()
        >>> theta = 0.123 * math.tau
        >>> components = {
        >>>     'rotate': kwimage.Affine.affine(theta=theta),
        >>>     'scale': kwimage.Affine.affine(scale=0.5),
        >>>     'shear': kwimage.Affine.affine(shearx=0.2),
        >>>     'translation': kwimage.Affine.affine(offset=(100, 200)),
        >>>     'rotate+translate': kwimage.Affine.affine(theta=0.123 * math.tau, about=(256, 256)),
        >>>     'random composed': kwimage.Affine.random(scale=(0.5, 1.5), translate=(-20, 20), theta=(-theta, theta), shearx=(0, .4), rng=900558176210808600),
        >>> }
        >>> warp_stack = []
        >>> for key, aff in components.items():
        ...     warp = kwimage.warp_affine(image, aff)
        ...     warp = kwimage.draw_text_on_image(
        ...        warp,
        ...        ub.urepr(aff.matrix, nl=1, nobr=1, precision=2, si=1, sv=1, with_dtype=0),
        ...        org=(1, 1),
        ...        valign='top', halign='left',
        ...        fontScale=0.8, color='kw_blue',
        ...        border={'thickness': 3},
        ...        )
        ...     warp = kwimage.draw_header_text(warp, key, color='kw_green')
        ...     warp_stack.append(warp)
        >>> warp_canvas = kwimage.stack_images_grid(warp_stack, chunksize=3, pad=10, bg_value='kitware_gray')
        >>> # xdoctest: +REQUIRES(module:sympy)
        >>> import sympy
        >>> # Shows the symbolic construction of the code
        >>> # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
        >>> from sympy.abc import theta
        >>> params = x0, y0, sx, sy, theta, shearx, tx, ty = sympy.symbols(
        >>>     'x0, y0, sx, sy, theta, shearx, tx, ty')
        >>> theta = sympy.symbols('theta')
        >>> # move the center to 0, 0
        >>> tr1_ = np.array([[1, 0,  -x0],
        >>>                  [0, 1,  -y0],
        >>>                  [0, 0,    1]])
        >>> # Define core components of the affine transform
        >>> S = np.array([  # scale
        >>>     [sx,  0, 0],
        >>>     [ 0, sy, 0],
        >>>     [ 0,  0, 1]])
        >>> E = np.array([  # x-shear
        >>>     [1,  shearx, 0],
        >>>     [0,  1, 0],
        >>>     [0,  0, 1]])
        >>> R = np.array([  # rotation
        >>>     [sympy.cos(theta), -sympy.sin(theta), 0],
        >>>     [sympy.sin(theta),  sympy.cos(theta), 0],
        >>>     [               0,                 0, 1]])
        >>> T = np.array([  # translation
        >>>     [ 1,  0, tx],
        >>>     [ 0,  1, ty],
        >>>     [ 0,  0,  1]])
        >>> # Contruct the affine 3x3 about the origin
        >>> aff0 = np.array(sympy.simplify(T @ R @ E @ S))
        >>> # move 0, 0 back to the specified origin
        >>> tr2_ = np.array([[1, 0,  x0],
        >>>                  [0, 1,  y0],
        >>>                  [0, 0,   1]])
        >>> # combine transformations
        >>> aff = tr2_ @ aff0 @ tr1_
        >>> print('aff = {}'.format(ub.urepr(aff.tolist(), nl=1)))
        >>> # This could be prettier
        >>> texts = {
        >>>     'Translation': sympy.pretty(sympy.Matrix(T), use_unicode=0),
        >>>     'Rotation': sympy.pretty(sympy.Matrix(R), use_unicode=0),
        >>>     'shEar-X': sympy.pretty(sympy.Matrix(E), use_unicode=0),
        >>>     'Scale': sympy.pretty(sympy.Matrix(S), use_unicode=0),
        >>> }
        >>> print(ub.urepr(texts, nl=2, sv=1))
        >>> equation_stack = []
        >>> for text, m in texts.items():
        >>>     render_canvas = kwimage.draw_text_on_image(None, m, color='kw_blue', fontScale=1.0)
        >>>     render_canvas = kwimage.draw_header_text(render_canvas, text, color='kw_green')
        >>>     render_canvas = kwimage.imresize(render_canvas, scale=1.3)
        >>>     equation_stack.append(render_canvas)
        >>> equation_canvas = kwimage.stack_images(equation_stack, pad=10, axis=1, bg_value='kitware_gray')
        >>> render_canvas = kwimage.draw_text_on_image(None, sympy.pretty(sympy.Array(aff), use_unicode=0), color='kw_blue', fontScale=1.0)
        >>> render_canvas = kwimage.draw_header_text(render_canvas, 'Full Equation With Pre-Shift', color='kw_green')
        >>> # xdoctest: -REQUIRES(module:sympy)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> plt = kwplot.autoplt()
        >>> canvas = kwimage.stack_images([warp_canvas, equation_canvas, render_canvas], pad=20, axis=0, bg_value='kitware_gray', resize='larger')
        >>> canvas = kwimage.draw_header_text(canvas, 'Affine matrixes can represent', color='kw_green')
        >>> kwplot.imshow(canvas)
        >>> fig = plt.gcf()
        >>> fig.set_size_inches(13, 13)

    Example:
        >>> import kwimage
        >>> self = kwimage.Affine(np.eye(3))
        >>> m1 = np.eye(3) @ self
        >>> m2 = self @ np.eye(3)

    Example:
        >>> from kwimage.transform import *  # NOQA
        >>> m = {}
        >>> # Works, and returns a Affine
        >>> m[len(m)] = x = Affine.random() @ np.eye(3)
        >>> assert isinstance(x, Affine)
        >>> m[len(m)] = x = Affine.random() @ None
        >>> assert isinstance(x, Affine)
        >>> # Works, and returns an ndarray
        >>> m[len(m)] = x = np.eye(3) @ Affine.random()
        >>> assert isinstance(x, np.ndarray)
        >>> # Works, and returns an Matrix
        >>> m[len(m)] = x = Affine.random() @ Matrix.random(3)
        >>> assert isinstance(x, Matrix)
        >>> m[len(m)] = x = Matrix.random(3) @ Affine.random()
        >>> assert isinstance(x, Matrix)
        >>> print('m = {}'.format(ub.urepr(m)))
    """
    @property
    def shape(self):
        return (3, 3)

    def __json__(self):
        if self.matrix is None:
            return {'type': 'affine', 'matrix': None}
        else:
            return {'type': 'affine', 'matrix': self.matrix.tolist()}

    def concise(self):
        """
        Return a concise coercable dictionary representation of this matrix

        Returns:
            Dict[str, object]: a small serializable dict that can be passed
                to :func:`Affine.coerce` to reconstruct this object.

        Returns:
            Dict: dictionary with consise parameters

        Example:
            >>> import kwimage
            >>> import numpy as np
            >>> import ubelt as ub
            >>> self = kwimage.Affine.random(rng=0, scale=1)
            >>> params = self.concise()
            >>> assert np.allclose(kwimage.Affine.coerce(params).matrix, self.matrix)
            >>> print('params = {}'.format(ub.urepr(params, nl=1, precision=2)))
            params = {
                'offset': (0.08, 0.38),
                'theta': 0.08,
                'type': 'affine',
            }

        Example:
            >>> import kwimage
            >>> import numpy as np
            >>> import ubelt as ub
            >>> self = kwimage.Affine.random(rng=0, scale=2, offset=0)
            >>> params = self.concise()
            >>> assert np.allclose(kwimage.Affine.coerce(params).matrix, self.matrix)
            >>> print('params = {}'.format(ub.urepr(params, nl=1, precision=2)))
            params = {
                'scale': 2.00,
                'theta': 0.04,
                'type': 'affine',
            }
        """
        params = self.decompose()
        params['type'] = 'affine'
        tx, ty = params['offset']
        sx, sy = params['scale']
        if math.isclose(tx, 0) and math.isclose(ty, 0):
            params.pop('offset')
        elif tx == ty:
            params['offset'] = tx
        if math.isclose(sy, 1) and math.isclose(sy, 1):
            params.pop('scale')
        elif sx == sy:
            params['scale'] = sx
        if math.isclose(params['shearx'], 0):
            params.pop('shearx')
        if math.isclose(params['theta'], 0):
            params.pop('theta')
        return params

    @classmethod
    def from_shapely(cls, sh_aff):
        """
        Shapely affine tuples are in the format (a, b, d, e, x, y)
        """
        (a, b, d, e, x, y) = sh_aff
        matrix = np.array([[a, b, x], [d, e, y], [0, 0, 1]])
        self = cls(matrix=matrix)
        return self

    @classmethod
    def from_affine(cls, aff):
        a, b, c, d, e, f = aff.a, aff.b, aff.c, aff.d, aff.e, aff.f
        matrix = np.array([[a, b, c], [d, e, f], [0, 0, 1]])
        self = cls(matrix=matrix)
        return self

    @classmethod
    def from_gdal(cls, gdal_aff):
        """
        gdal affine tuples are in the format (c, a, b, f, d, e)
        """
        c, a, b, f, d, e = gdal_aff
        matrix = np.array([[a, b, c], [d, e, f], [0, 0, 1]])
        self = cls(matrix=matrix)
        return self

    @classmethod
    def from_skimage(cls, sk_aff):
        """
        gdal affine tuples are in the format (c, a, b, f, d, e)
        """
        self = cls(matrix=sk_aff.params)
        return self

    @classmethod
    def coerce(cls, data=None, **kwargs):
        """
        Attempt to coerce the data into an affine object

        Args:
            data : some data we attempt to coerce to an Affine matrix
            **kwargs : some data we attempt to coerce to an Affine matrix,
                mutually exclusive with `data`.

        Returns:
            Affine

        Example:
            >>> import kwimage
            >>> import skimage.transform
            >>> kwimage.Affine.coerce({'type': 'affine', 'matrix': [[1, 0, 0], [0, 1, 0]]})
            >>> kwimage.Affine.coerce({'scale': 2})
            >>> kwimage.Affine.coerce({'offset': 3})
            >>> kwimage.Affine.coerce(np.eye(3))
            >>> kwimage.Affine.coerce(None)
            >>> kwimage.Affine.coerce({})
            >>> kwimage.Affine.coerce(skimage.transform.AffineTransform(scale=30))
        """
        if data is None and not kwargs:
            # Just use a real eye matrix here.
            return cls(matrix=None)
        if data is None:
            data = kwargs
        if isinstance(data, np.ndarray):
            self = cls(matrix=data)
        elif isinstance(data, cls):
            self = data
        elif isinstance(data, skimage.transform.AffineTransform):
            self = cls(matrix=data.params)
        elif data.__class__.__name__ == cls.__name__:
            self = data
        elif isinstance(data, tuple):
            raise ValueError(
                'Cannot determine if a tuple is in shapely or gdal order.'
                'use from_shapely or from_gdal instead'
            )
        elif isinstance(data, dict):
            keys = set(data.keys())
            if 'matrix' in keys:
                self = cls(matrix=np.array(data['matrix']))
            else:
                known_params = {'scale', 'offset', 'theta', 'type', 'shearx', 'shear', 'about'}
                params = {key: data[key] for key in known_params if key in data}
                unknown_params = keys - known_params
                if len(unknown_params) == 0:
                    params.pop('type', None)
                    _nkeys = len(keys)
                    if _nkeys == 1:
                        # Special cases for speed
                        if keys == {'scale'}:
                            self = cls.scale(**params)
                        elif keys == {'offset'}:
                            self = cls.translate(**params)
                        else:
                            self = cls.affine(**params)
                    # elif _nkeys == 2:  # may not be worth it
                    #     # Special cases for speed
                    #     if keys == {'scale', 'offset'}:
                    #         self = cls._scale_translate(**params)
                    #     else:
                    #         self = cls.affine(**params)
                    else:
                        self = cls.affine(**params)
                else:
                    got_known_parms = set(params) - unknown_params
                    raise KeyError(
                        'Got known params: ' + ', '.join(list(got_known_parms)) + ' '
                        'Got unknown params: ' + ', '.join(list(unknown_params)))
        else:
            raise TypeError(type(data))
        return self

    def eccentricity(self):
        """
        Eccentricity of the ellipse formed by this affine matrix

        Returns:
            float: large when there are big scale differences in principle
                directions or skews.

        References:
            .. [WikiConic] https://en.wikipedia.org/wiki/Conic_section
            .. [GHAffine] https://github.com/rasterio/affine/blob/78c20a0cfbb5ec/affine/__init__.py#L368

        Example:
            >>> import kwimage
            >>> kwimage.Affine.random(rng=432).eccentricity()
        """
        # Ignore the translation part
        M = self.matrix[0:2, 0:2]

        MMt = M @ M.T
        trace = np.trace(MMt)
        det = np.linalg.det(MMt)

        root_delta = np.sqrt((trace * trace) / 4 - det)

        # scaling defined via affine.Affine.
        ell1 = np.sqrt(trace / 2 + root_delta)
        ell2 = np.sqrt(trace / 2 - root_delta)

        ecc = np.sqrt(ell1 * ell1 - ell2 * ell2) / ell1
        return ecc

    def to_affine(self):
        """
        Convert to an affine module

        Returns:
            affine.Affine
        """
        import affine
        a, b, c, d, e, f = self.matrix.ravel()[0:6]
        aff = affine.Affine(a, b, c, d, e, f)
        return aff

    def to_gdal(self):
        """
        Convert to a gdal tuple (c, a, b, f, d, e)

        Returns:
            Tuple[float, float, float, float, float, float]
        """
        return self.to_affine().to_gdal()

    def to_shapely(self):
        """
        Returns a matrix suitable for shapely.affinity.affine_transform

        Returns:
            Tuple[float, float, float, float, float, float]

        Example:
            >>> import kwimage
            >>> self = kwimage.Affine.random()
            >>> sh_transform = self.to_shapely()
            >>> # Transform points with kwimage and shapley
            >>> import shapely
            >>> from shapely.affinity import affine_transform
            >>> kw_poly = kwimage.Polygon.random()
            >>> kw_warp_poly = kw_poly.warp(self)
            >>> sh_poly = kw_poly.to_shapely()
            >>> sh_warp_poly = affine_transform(sh_poly, sh_transform)
            >>> kw_warp_poly_recon = kwimage.Polygon.from_shapely(sh_warp_poly)
            >>> assert np.allclose(kw_warp_poly_recon.exterior.data, kw_warp_poly_recon.exterior.data)
        """
        # from shapely.affinity import affine_transform
        a, b, x, d, e, y = self.matrix.ravel()[0:6]
        sh_transform = (a, b, d, e, x, y)
        return sh_transform

    def to_skimage(self):
        """
        Returns:
            skimage.transform.AffineTransform

        Example:
            >>> import kwimage
            >>> self = kwimage.Affine.random()
            >>> tf = self.to_skimage()
            >>> # Transform points with kwimage and scikit-image
            >>> kw_poly = kwimage.Polygon.random()
            >>> kw_warp_xy = kw_poly.warp(self.matrix).exterior.data
            >>> sk_warp_xy = tf(kw_poly.exterior.data)
            >>> assert np.allclose(sk_warp_xy, sk_warp_xy)
        """
        return skimage.transform.AffineTransform(matrix=np.asarray(self))

    @classmethod
    def scale(cls, scale):
        """
        Create a scale Affine object

        Args:
            scale (float | Tuple[float, float]): x, y scale factor

        Returns:
            Affine
        """
        scale_ = 1 if scale is None else scale
        sx, sy = _ensure_iterable2(scale_)
        # Sympy simplified expression
        mat = np.array([sx , 0.0, 0.0,
                        0.0,  sy, 0.0,
                        0.0, 0.0, 1.0])
        mat = mat.reshape(3, 3)  # Faster to make a flat array and reshape
        self = cls(mat)
        return self

    @classmethod
    def translate(cls, offset):
        """
        Create a translation Affine object

        Args:
            offset (float | Tuple[float, float]): x, y translation factor

        Returns:
            Affine

        Benchmark:
            >>> # xdoctest: +REQUIRES(--benchmark)
            >>> # It is ~3x faster to use the more specific method
            >>> import timerit
            >>> import kwimage
            >>> #
            >>> offset = np.random.rand(2)
            >>> ti = timerit.Timerit(100, bestof=10, verbose=2)
            >>> for timer in ti.reset('time'):
            >>>     with timer:
            >>>         kwimage.Affine.translate(offset)
            >>> #
            >>> for timer in ti.reset('time'):
            >>>     with timer:
            >>>         kwimage.Affine.affine(offset=offset)

        """
        offset_ = 0 if offset is None else offset
        tx, ty = _ensure_iterable2(offset_)
        # Sympy simplified expression
        mat = np.array([1.0, 0.0, tx,
                        0.0, 1.0, ty,
                        0.0, 0.0, 1.0])
        mat = mat.reshape(3, 3)  # Faster to make a flat array and reshape
        self = cls(mat)
        return self

    @classmethod
    def _scale_translate(cls, scale, offset):
        """ helper method for speed """
        scale_ = 1 if scale is None else scale
        offset_ = 0 if offset is None else offset
        sx, sy = _ensure_iterable2(scale_)
        tx, ty = _ensure_iterable2(offset_)
        # Sympy simplified expression
        mat = np.array([sx , 0.0, tx,
                        0.0,  sy, ty,
                        0.0, 0.0, 1.0])
        mat = mat.reshape(3, 3)  # Faster to make a flat array and reshape
        self = cls(mat)
        return self

    @classmethod
    def rotate(cls, theta):
        """
        Create a rotation Affine object

        Args:
            theta (float): counter-clockwise rotation angle in radians

        Returns:
            Affine
        """
        return cls.affine(theta=theta)

    @classmethod
    def random(cls, shape=None, rng=None, **kw):
        """
        Create a random Affine object

        Args:
            rng : random number generator
            **kw: passed to :func:`Affine.random_params`.
                can contain coercable random distributions for scale, offset,
                about, theta, and shearx.

        Returns:
            Affine
        """
        if shape is not None:
            raise ValueError('cannot specify shape to Affine.random')
        params = cls.random_params(rng=rng, **kw)
        self = cls.affine(**params)
        return self

    @classmethod
    def random_params(cls, rng=None, **kw):
        """
        Args:
            rng : random number generator
            **kw: can contain coercable random distributions for
                scale, offset, about, theta, and shearx.

        Returns:
            Dict: affine parameters suitable to be passed to Affine.affine

        TODO:
            - [ ] improve kwargs parameterization
        """
        from kwarray import distributions
        import numbers
        TN = distributions.TruncNormal
        rng = kwarray.ensure_rng(rng)

        def _coerce_distri(arg):
            if isinstance(arg, numbers.Number):
                dist = distributions.Constant(arg, rng=rng)
            elif isinstance(arg, tuple) and len(arg) == 2:
                lo, hi = arg
                dist = distributions.Uniform(low=lo, high=hi, rng=rng)
            else:
                raise NotImplementedError
            return dist

        if 'scale' in kw:
            if ub.iterable(kw['scale']) and (not isinstance(kw['scale'], tuple) and len(kw['scale']) == 2):
                raise NotImplementedError
            else:
                xscale_dist = _coerce_distri(kw['scale'])
                yscale_dist = xscale_dist
        else:
            scale_kw = dict(mean=1, std=1, low=1, high=2)
            xscale_dist = TN(**scale_kw, rng=rng)
            yscale_dist = TN(**scale_kw, rng=rng)

        if 'offset' in kw:
            if ub.iterable(kw['offset']):
                raise NotImplementedError
            else:
                xoffset_dist = _coerce_distri(kw['offset'])
                yoffset_dist = xoffset_dist
        else:
            offset_kw = dict(mean=0, std=1, low=-1, high=1)
            xoffset_dist = TN(**offset_kw, rng=rng)
            yoffset_dist = TN(**offset_kw, rng=rng)

        if 'about' in kw:
            if ub.iterable(kw['about']):
                raise NotImplementedError
            else:
                xabout_dist = _coerce_distri(kw['about'])
                yabout_dist = xabout_dist
        else:
            xabout_dist = distributions.Constant(0, rng=rng)
            yabout_dist = distributions.Constant(0, rng=rng)

        if 'theta' in kw:
            theta_dist = _coerce_distri(kw['theta'])
        else:
            theta_kw = dict(mean=0, std=1, low=-np.pi / 8, high=np.pi / 8)
            theta_dist = TN(**theta_kw, rng=rng)

        if 'shearx' in kw:
            shear_dist = _coerce_distri(kw['shearx'])
        else:
            shear_dist = distributions.Constant(0, rng=rng)

        # scale_kw = dict(mean=1, std=1, low=0, high=2)
        # offset_kw = dict(mean=0, std=1, low=-1, high=1)
        # theta_kw = dict(mean=0, std=1, low=-6.28, high=6.28)

        # TODO: distributions.Distribution.coerce()
        # offset_dist = distributions.Constant(0)
        # theta_dist = distributions.Constant(0)

        # todo better parametarization
        params = dict(
            scale=(xscale_dist.sample(), yscale_dist.sample()),
            offset=(xoffset_dist.sample(), yoffset_dist.sample()),
            theta=theta_dist.sample(),
            shearx=shear_dist.sample(),
            about=(xabout_dist.sample(), yabout_dist.sample()),
        )
        return params

    @profile
    def decompose(self):
        r"""
        Decompose the affine matrix into its individual scale, translation,
        rotation, and skew parameters.

        Returns:
            Dict: decomposed offset, scale, theta, and shearx params

        References:
            .. [SE612006] https://math.stackexchange.com/questions/612006/decompose-affine
            .. [SE3521141] https://math.stackexchange.com/a/3521141/353527
            .. [SE70357473] https://stackoverflow.com/questions/70357473/how-to-decompose-a-2x2-affine-matrix-with-sympy
            .. [WikiTranMat] https://en.wikipedia.org/wiki/Transformation_matrix
            .. [WikiShear] https://en.wikipedia.org/wiki/Shear_mapping

        Example:
            >>> from kwimage.transform import *  # NOQA
            >>> self = Affine.random()
            >>> params = self.decompose()
            >>> recon = Affine.coerce(**params)
            >>> params2 = recon.decompose()
            >>> pt = np.vstack([np.random.rand(2, 1), [1]])
            >>> result1 = self.matrix[0:2] @ pt
            >>> result2 = recon.matrix[0:2] @ pt
            >>> assert np.allclose(result1, result2)

            >>> self = Affine.scale(0.001) @ Affine.random()
            >>> params = self.decompose()
            >>> self.det()

        Example:
            >>> # xdoctest: +REQUIRES(module:sympy)
            >>> # Test decompose with symbolic matrices
            >>> from kwimage.transform import *  # NOQA
            >>> self = Affine.random().rationalize()
            >>> self.decompose()

        Example:
            >>> # xdoctest: +REQUIRES(module:pandas)
            >>> from kwimage.transform import *  # NOQA
            >>> import kwimage
            >>> import pandas as pd
            >>> # Test consistency of decompose + reconstruct
            >>> param_grid = list(ub.named_product({
            >>>     'theta': np.linspace(-4 * np.pi, 4 * np.pi, 3),
            >>>     'shearx': np.linspace(- 10 * np.pi, 10 * np.pi, 4),
            >>> }))
            >>> def normalize_angle(radian):
            >>>     return np.arctan2(np.sin(radian), np.cos(radian))
            >>> for pextra in param_grid:
            >>>     params0 = dict(scale=(3.05, 3.07), offset=(10.5, 12.1), **pextra)
            >>>     self = recon0 = kwimage.Affine.affine(**params0)
            >>>     self.decompose()
            >>>     # Test drift with multiple decompose / reconstructions
            >>>     params_list = [params0]
            >>>     recon_list = [recon0]
            >>>     n = 4
            >>>     for _ in range(n):
            >>>         prev = recon_list[-1]
            >>>         params = prev.decompose()
            >>>         recon = kwimage.Affine.coerce(**params)
            >>>         params_list.append(params)
            >>>         recon_list.append(recon)
            >>>     params_df = pd.DataFrame(params_list)
            >>>     #print('params_list = {}'.format(ub.urepr(params_list, nl=1, precision=5)))
            >>>     print(params_df)
            >>>     assert ub.allsame(normalize_angle(params_df['theta']), eq=np.isclose)
            >>>     assert ub.allsame(params_df['shearx'], eq=np.allclose)
            >>>     assert ub.allsame(params_df['scale'], eq=np.allclose)
            >>>     assert ub.allsame(params_df['offset'], eq=np.allclose)

        Ignore:
            import affine
            self = Affine.random()
            a, b, c, d, e, f = self.matrix.ravel()[0:6]
            aff = affine.Affine(a, b, c, d, e, f)
            assert np.isclose(self.det(), aff.determinant)

            params = self.decompose()
            assert np.isclose(params['theta'], np.deg2rad(aff.rotation_angle))

            print(params['scale'])
            print(aff._scaling)
            print(self.eccentricity())
            print(aff.eccentricity)

        Ignore:
            import timerit
            ti = timerit.Timerit(100, bestof=10, verbose=2)
            for timer in ti.reset('time'):
                self = Affine.random()
                with timer:
                    self.decompose()

            # Wow: using math instead of numpy for scalars is much faster!

            for timer in ti.reset('time'):
                with timer:
                    math.sqrt(a11 * a11 + a21 * a21)

            for timer in ti.reset('time'):
                with timer:
                    np.arctan2(a21, a11)

            for timer in ti.reset('time'):
                with timer:
                    math.atan2(a21, a11)
        """
        if self.matrix is None:
            return {'offset': (0., 0.), 'scale': (1., 1.), 'shearx': 0.,
                    'theta': 0., }
        a11, a12, a13, a21, a22, a23 = self.matrix.ravel()[0:6]

        if self.is_rational():
            math_mod = sympy
        else:
            math_mod = math

        sx = math_mod.sqrt(a11 * a11 + a21 * a21)
        theta = math_mod.atan2(a21, a11)
        sin_t = math_mod.sin(theta)
        cos_t = math_mod.cos(theta)

        msy = a12 * cos_t + a22 * sin_t

        try:
            if abs(cos_t) < abs(sin_t):
                sy = (msy * cos_t - a12) / sin_t
            else:
                sy = (a22 - msy * sin_t) / cos_t
        except TypeError:
            # symbolic issue
            sy = sympy.Piecewise(
                ((msy * cos_t - a12) / sin_t, abs(cos_t) < abs(sin_t)),
                ((a22 - msy * sin_t) / cos_t, True)
            )

        shearx = msy / sy
        tx, ty = a13, a23

        params = {
            'offset': (tx, ty),
            'scale': (sx, sy),
            'shearx': shearx,
            'theta': theta,
        }
        return params

    @profile
    def _decompose_scale(self):
        """
        Scale only decomposition. Experimental method that is faster than
        decompose when only scale is needed.
        """
        if self.matrix is None:
            return (1., 1.)
        a11, a12, _, a21, a22 = self.matrix.ravel()[0:5]
        sx = math.sqrt(a11 * a11 + a21 * a21)
        theta = math.atan2(a21, a11)
        sin_t = math.sin(theta)
        cos_t = math.cos(theta)
        msy = a12 * cos_t + a22 * sin_t
        if abs(cos_t) < abs(sin_t):
            sy = (msy * cos_t - a12) / sin_t
        else:
            sy = (a22 - msy * sin_t) / cos_t
        scale = (sx, sy)
        return scale

    @classmethod
    def affine(cls, scale=None, offset=None, theta=None, shear=None,
               about=None, shearx=None, array_cls=None, math_mod=None,
               **kwargs):
        """
        Create an affine matrix from high-level parameters

        Args:
            scale (float | Tuple[float, float]):
                x, y scale factor

            offset (float | Tuple[float, float]):
                x, y translation factor

            theta (float):
                counter-clockwise rotation angle in radians

            shearx (float):
                shear factor parallel to the x-axis.

            about (float | Tuple[float, float]):
                x, y location of the origin

            shear (float):
                BROKEN, dont use.  counter-clockwise shear angle in radians

        TODO:
            - [ ] Add aliases? -
                origin : alias for about
                rotation : alias for theta
                translation : alias for offset

        Returns:
            Affine: the constructed Affine object

        Example:
            >>> from kwimage.transform import *  # NOQA
            >>> rng = kwarray.ensure_rng(None)
            >>> scale = rng.randn(2) * 10
            >>> offset = rng.randn(2) * 10
            >>> about = rng.randn(2) * 10
            >>> theta = rng.randn() * 10
            >>> shearx = rng.randn() * 10
            >>> # Create combined matrix from all params
            >>> F = Affine.affine(
            >>>     scale=scale, offset=offset, theta=theta, shearx=shearx,
            >>>     about=about)
            >>> # Test that combining components matches
            >>> S = Affine.affine(scale=scale)
            >>> T = Affine.affine(offset=offset)
            >>> R = Affine.affine(theta=theta)
            >>> E = Affine.affine(shearx=shearx)
            >>> O = Affine.affine(offset=about)
            >>> # combine (note shear must be on the RHS of rotation)
            >>> alt  = O @ T @ R @ E @ S @ O.inv()
            >>> print('F    = {}'.format(ub.urepr(F.matrix.tolist(), nl=1)))
            >>> print('alt  = {}'.format(ub.urepr(alt.matrix.tolist(), nl=1)))
            >>> assert np.all(np.isclose(alt.matrix, F.matrix))
            >>> pt = np.vstack([np.random.rand(2, 1), [[1]]])
            >>> warp_pt1 = (F.matrix @ pt)
            >>> warp_pt2 = (alt.matrix @ pt)
            >>> assert np.allclose(warp_pt2, warp_pt1)

        Sympy:
            >>> # xdoctest: +SKIP
            >>> import sympy
            >>> # Shows the symbolic construction of the code
            >>> # https://groups.google.com/forum/#!topic/sympy/k1HnZK_bNNA
            >>> from sympy.abc import theta
            >>> params = x0, y0, sx, sy, theta, shearx, tx, ty = sympy.symbols(
            >>>     'x0, y0, sx, sy, theta, shearx, tx, ty')
            >>> # move the center to 0, 0
            >>> tr1_ = np.array([[1, 0,  -x0],
            >>>                  [0, 1,  -y0],
            >>>                  [0, 0,    1]])
            >>> # Define core components of the affine transform
            >>> S = np.array([  # scale
            >>>     [sx,  0, 0],
            >>>     [ 0, sy, 0],
            >>>     [ 0,  0, 1]])
            >>> E = np.array([  # x-shear
            >>>     [1,  shearx, 0],
            >>>     [0,  1, 0],
            >>>     [0,  0, 1]])
            >>> R = np.array([  # rotation
            >>>     [sympy.cos(theta), -sympy.sin(theta), 0],
            >>>     [sympy.sin(theta),  sympy.cos(theta), 0],
            >>>     [               0,                 0, 1]])
            >>> T = np.array([  # translation
            >>>     [ 1,  0, tx],
            >>>     [ 0,  1, ty],
            >>>     [ 0,  0,  1]])
            >>> # Contruct the affine 3x3 about the origin
            >>> aff0 = np.array(sympy.simplify(T @ R @ E @ S))
            >>> # move 0, 0 back to the specified origin
            >>> tr2_ = np.array([[1, 0,  x0],
            >>>                  [0, 1,  y0],
            >>>                  [0, 0,   1]])
            >>> # combine transformations
            >>> aff = tr2_ @ aff0 @ tr1_
            >>> print('aff = {}'.format(ub.urepr(aff.tolist(), nl=1)))

        Ignore:
            import timerit
            ti = timerit.Timerit(10000, bestof=10, verbose=2)
            for timer in ti.reset('time'):
                with timer:
                    self = kwimage.Affine.affine(scale=3, offset=2, theta=np.random.rand(), shearx=np.random.rand())
        """
        if shear is not None and shearx is None:
            # Hack so old data is readable (this should be ok as long as the
            # data wasnt reserialized)
            if not _internal.KWIMAGE_DISABLE_TRANSFORM_WARNINGS:
                ub.schedule_deprecation(
                    modname='kwimage', name='shear', type='parameter',
                    migration=ub.paragraph(
                        '''
                        The `shear` parameter is deprecated and will be removed because
                        of a serious bug. Use `shearx` instead. See Issue #8 on
                        https://gitlab.kitware.com/computer-vision/kwimage/-/issues/8
                        for more details. To ease the impact of this bug we will
                        interpret `shear` as `shearx`, which should result in a correct
                        reconstruction, as long as the data was never reserialized.
                        '''
                    ), deprecate='0.9.0', error='0.10.0', remove='0.11.0', warncls=UserWarning)
            shearx = shear
            shear = None

        if array_cls is None:
            array_cls = np.array

        if math_mod is None:
            math_mod = math

        scale_ = 1 if scale is None else scale
        offset_ = 0 if offset is None else offset
        xshear_ = 0 if shearx is None else shearx
        theta_ = 0 if theta is None else theta
        about_ = 0 if about is None else about
        sx, sy = _ensure_iterable2(scale_)
        tx, ty = _ensure_iterable2(offset_)
        x0, y0 = _ensure_iterable2(about_)

        cos_theta = math_mod.cos(theta_)
        sin_theta = math_mod.sin(theta_)

        sx_cos_theta = sx * cos_theta
        sx_sin_theta = sx * sin_theta

        sy_cos_theta = sy * cos_theta
        sy_sin_theta = sy * sin_theta

        m_sy_cos_theta = xshear_ * sy_cos_theta
        m_sy_sin_theta = xshear_ * sy_sin_theta

        a12 = m_sy_cos_theta - sy_sin_theta
        a22 = m_sy_sin_theta + sy_cos_theta

        tx_ = tx + x0 - (x0 * sx_cos_theta) - (y0 * a12)
        ty_ = ty + y0 - (x0 * sx_sin_theta) - (y0 * a22)

        mat = array_cls([sx_cos_theta, a12, tx_,
                         sx_sin_theta, a22, ty_,
                                    0,   0,  1])
        mat = mat.reshape(3, 3)  # Faster to make a flat array and reshape
        self = cls(mat)
        return self

    @classmethod
    def fit(cls, pts1, pts2):
        """
        Fit an affine transformation between a set of corresponding points

        Args:
            pts1 (ndarray): An Nx2 array of points in "space 1".
            pts2 (ndarray): A corresponding Nx2 array of points in "space 2"

        Returns:
            Affine : a transform that warps from "space1" to "space2".

        Note:
            An affine matrix has 6 degrees of freedom, so at least 3
            non-colinear xy-point pairs are needed.

        References:
            ..[Lowe04] https://www.cs.ubc.ca/~lowe/papers/ijcv04.pdf page 22

        Example:
            >>> # Create a set of points, warp them, then recover the warp
            >>> import kwimage
            >>> points = kwimage.Points.random(6).scale(64)
            >>> #A1 = kwimage.Affine.affine(scale=0.9, theta=-3.2, offset=(2, 3), about=(32, 32), skew=2.3)
            >>> #A2 = kwimage.Affine.affine(scale=0.8, theta=0.8, offset=(2, 0), about=(32, 32))
            >>> A1 = kwimage.Affine.random()
            >>> A2 = kwimage.Affine.random()
            >>> A12_real = A2 @ A1.inv()
            >>> points1 = points.warp(A1)
            >>> points2 = points.warp(A2)
            >>> # Recover the warp
            >>> pts1, pts2 = points1.xy, points2.xy
            >>> A_recovered = kwimage.Affine.fit(pts1, pts2)
            >>> assert np.all(np.isclose(A_recovered.matrix, A12_real.matrix))
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> base1 = np.zeros((96, 96, 3))
            >>> base1[32:-32, 5:-5] = 0.5
            >>> base2 = np.zeros((96, 96, 3))
            >>> img1 = points1.draw_on(base1, radius=3, color='blue')
            >>> img2 = points2.draw_on(base2, radius=3, color='green')
            >>> img1_warp = kwimage.warp_affine(img1, A_recovered)
            >>> canvas = kwimage.stack_images([img1, img2, img1_warp], pad=10, axis=1, bg_value=(1., 1., 1.))
            >>> kwplot.imshow(canvas)
        """
        if 0:
            # Not sure if cv2 has this variant of the affine matrix calc
            import cv2
            inlier_method = 'ransac'
            inlier_method_lut = {
                'lmeds': cv2.LMEDS,
                'ransac': cv2.RANSAC,
            }
            cv2_method = inlier_method_lut[inlier_method]
            A, mask = cv2.estimateAffine2D(pts1, pts2, method=cv2_method)

        num_pts = len(pts1)
        if num_pts == 0:
            # Regularize degenerate 0 correspondence case
            mat = np.eye(3)  # Identity if no points
        elif num_pts == 1:
            # Regularize degenerate 1 correspondence case
            # Only translation possible
            t1 = np.mean(pts1, axis=0)
            t2 = np.mean(pts2, axis=0)
            tx, ty = t2 - t1
            mat = np.array([
                [1, 0, tx],
                [0, 1, ty],
                [0, 0, 1]
            ])
        elif num_pts == 2:
            # Regularize degenerate 2 correspondence case
            # For 2 points, compute rotation and scaling
            # Center points to estimate rotation/scale
            t1 = np.mean(pts1, axis=0)
            t2 = np.mean(pts2, axis=0)
            centered1 = pts1 - t1
            centered2 = pts2 - t2

            # Solve for rotation and scaling using Kabsch algorithm (simplified)
            # https://en.wikipedia.org/wiki/Kabsch_algorithm
            H = centered1.T @ centered2  # Cross-covariance matrix
            U, _, Vt = np.linalg.svd(H)
            R = Vt.T @ U.T  # Rotation matrix

            # Handle reflection case (enforce proper rotation)
            if np.linalg.det(R) < 0:
                Vt[-1, :] *= -1
                R = Vt.T @ U.T

            # Estimate scaling
            s = np.trace(H @ R) / np.trace(centered1.T @ centered1)

            # Build full transform: [sR | t2 - sR t1]
            mat = np.eye(3)
            mat[:2, :2] = s * R
            mat[:2, 2] = t2 - (s * R) @ t1
        else:
            x1_mn = pts1[:, 0]
            y1_mn = pts1[:, 1]
            x2_mn = pts2[:, 0]
            y2_mn = pts2[:, 1]
            num_pts = x1_mn.shape[0]
            Mx6 = np.empty((2 * num_pts, 6), dtype=float)
            b = np.empty((2 * num_pts, 1), dtype=float)
            for ix in range(num_pts):  # Loop over inliers
                # Concatenate all 2x9 matrices into an Mx6 matrix
                x1 = x1_mn[ix]
                x2 = x2_mn[ix]
                y1 = y1_mn[ix]
                y2 = y2_mn[ix]
                Mx6[ix * 2]     = (x1, y1, 0, 0, 1, 0)
                Mx6[ix * 2 + 1] = ( 0, 0, x1, y1, 0, 1)
                b[ix * 2] = x2
                b[ix * 2 + 1] = y2

            M = Mx6
            try:
                USVt = np.linalg.svd(M, full_matrices=True, compute_uv=True)
            except MemoryError:
                import scipy.sparse as sps
                import scipy.sparse.linalg as spsl
                M_sparse = sps.lil_matrix(M)
                USVt = spsl.svds(M_sparse)
            except np.linalg.LinAlgError:
                raise
            except Exception:
                raise

            U, s, Vt = USVt

            if np.isclose(s[-1], 0):
                # SVD solution failed, fallback on lstsq
                a, _, _, _ = np.linalg.lstsq(M, b, rcond=None)
                a = a.ravel()
            else:
                # Inefficient, but the math works
                # We want to solve Ax=b (where A is the Mx6 in this case)
                # Ax = b
                # (U S V.T) x = b
                # x = (U.T inv(S) V) b
                Sinv = np.zeros((len(Vt), len(U)))
                Sinv[np.diag_indices(len(s))] = 1 / s
                a = Vt.T.dot(Sinv).dot(U.T).dot(b).T[0]
            mat = np.array([
                [a[0], a[1], a[4]],
                [a[2], a[3], a[5]],
                [   0, 0, 1],
            ])
        return Affine(mat)

    @classmethod
    def fliprot(cls, flip_axis=None, rot_k=0, axes=(0, 1), canvas_dsize=None):
        """
        Creates a flip/rotation transform with respect to an image of a given
        size in the positive quadrent. (i.e. warped data within the specified
        canvas size will stay in the positive quadrant)

        Args:
            flip_axis (int): the axis dimension to flip.
                I.e. 0 flips the y-axis and 1-flips the x-axis.

            rot_k (int): number of counterclockwise 90 degree rotations that
                occur after the flips.

            axes (Tuple[int, int]):
                The axis ordering. Unhandled in this version. Dont change this.

            canvas_dsize (Tuple[int, int]):
                The width / height of the canvas the fliprot is applied in.

        Returns:
            Affine:
                The affine matrix representing the canvas-aligned flip and
                rotation.

        Note:
            Requiring that the image size is known makes this a place that
            errors could occur depending on your interpretation of pixels as
            points or areas. There is probably a better way to describe the
            issue, but the second doctest shows the issue when trying to use
            warp-affine's auto-dsize feature. See [MR81]_ for details.

        References:
            .. [SO57863376] https://stackoverflow.com/questions/57863376/flip-image-affine
            .. [MR81] https://gitlab.kitware.com/computer-vision/kwimage/-/merge_requests/81

        CommandLine:
            xdoctest -m kwimage.transform Affine.fliprot:0 --show
            xdoctest -m kwimage.transform Affine.fliprot:1 --show

        Example:
            >>> import kwimage
            >>> H, W = 64, 128
            >>> canvas_dsize = (W, H)
            >>> box1 = kwimage.Boxes.random(1).scale((W, H)).quantize()
            >>> ltrb = box1.data
            >>> rot_k = 4
            >>> annot = box1
            >>> annot = box1.to_polygons()[0]
            >>> annot1 = annot.copy()
            >>> # The first 8 are the cannonically unique group elements
            >>> fliprot_params = [
            >>>     {'rot_k': 0, 'flip_axis': None},
            >>>     {'rot_k': 1, 'flip_axis': None},
            >>>     {'rot_k': 2, 'flip_axis': None},
            >>>     {'rot_k': 3, 'flip_axis': None},
            >>>     {'rot_k': 0, 'flip_axis': (0,)},
            >>>     {'rot_k': 1, 'flip_axis': (0,)},
            >>>     {'rot_k': 2, 'flip_axis': (0,)},
            >>>     {'rot_k': 3, 'flip_axis': (0,)},
            >>>     # The rest of these dont result in any different data, but we need to test them
            >>>     {'rot_k': 0, 'flip_axis': (1,)},
            >>>     {'rot_k': 1, 'flip_axis': (1,)},
            >>>     {'rot_k': 2, 'flip_axis': (1,)},
            >>>     {'rot_k': 3, 'flip_axis': (1,)},
            >>>     {'rot_k': 0, 'flip_axis': (0, 1)},
            >>>     {'rot_k': 1, 'flip_axis': (0, 1)},
            >>>     {'rot_k': 2, 'flip_axis': (0, 1)},
            >>>     {'rot_k': 3, 'flip_axis': (0, 1)},
            >>> ]
            >>> results = []
            >>> for params in fliprot_params:
            >>>     tf = kwimage.Affine.fliprot(canvas_dsize=canvas_dsize, **params)
            >>>     annot2 = annot.warp(tf)
            >>>     annot3 = annot2.warp(tf.inv())
            >>>     #annot3 = inv_fliprot_annot(annot2, canvas_dsize=canvas_dsize, **params)
            >>>     results.append({
            >>>         'annot2': annot2,
            >>>         'annot3': annot3,
            >>>         'params': params,
            >>>         'tf': tf,
            >>>         'canvas_dsize': canvas_dsize,
            >>>     })
            >>> box = kwimage.Box.coerce([0, 0, W, H], format='xywh')
            >>> for result in results:
            >>>     params = result['params']
            >>>     warped = box.warp(result['tf'])
            >>>     print('---')
            >>>     print('params = {}'.format(ub.urepr(params, nl=1)))
            >>>     print('box = {}'.format(ub.urepr(box, nl=1)))
            >>>     print('warped = {}'.format(ub.urepr(warped, nl=1)))
            >>>     print(ub.hzcat(['tf = ', ub.urepr(result['tf'], nl=1)]))

            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> S = max(W, H)
            >>> image1 = kwimage.grab_test_image('astro', dsize=(S, S))[:H, :W]
            >>> pnum_ = kwplot.PlotNums(nCols=4, nSubplots=len(results))
            >>> for result in results:
            >>>     #image2 = kwimage.warp_affine(image1.copy(), result['tf'], dsize=(S, S))  # fixme dsize=positive should work here
            >>>     image2 = kwimage.warp_affine(image1.copy(), result['tf'], dsize='positive')  # fixme dsize=positive should work here
            >>>     #image3 = kwimage.warp_affine(image2.copy(), result['tf'].inv(), dsize=(S, S))
            >>>     image3 = kwimage.warp_affine(image2.copy(), result['tf'].inv(), dsize='positive')
            >>>     annot2 = result['annot2']
            >>>     annot3 = result['annot3']
            >>>     canvas1 = annot1.draw_on(image1.copy(), edgecolor='kitware_blue', fill=False)
            >>>     canvas2 = annot2.draw_on(image2.copy(), edgecolor='kitware_green', fill=False)
            >>>     canvas3 = annot3.draw_on(image3.copy(), edgecolor='kitware_red', fill=False)
            >>>     canvas = kwimage.stack_images([canvas1, canvas2, canvas3], axis=1, pad=10, bg_value='green')
            >>>     kwplot.imshow(canvas, pnum=pnum_(), title=ub.urepr(result['params'], nl=0, compact=1, nobr=1))
            >>> kwplot.show_if_requested()

        Example:
            >>> # Second similar test with a very small image to catch small errors
            >>> import kwimage
            >>> H, W = 4, 8
            >>> canvas_dsize = (W, H)
            >>> box1 = kwimage.Boxes.random(1).scale((W, H)).quantize()
            >>> ltrb = box1.data
            >>> rot_k = 4
            >>> annot = box1
            >>> annot = box1.to_polygons()[0]
            >>> annot1 = annot.copy()
            >>> # The first 8 are the cannonically unique group elements
            >>> fliprot_params = [
            >>>     {'rot_k': 0, 'flip_axis': None},
            >>>     {'rot_k': 1, 'flip_axis': None},
            >>>     {'rot_k': 2, 'flip_axis': None},
            >>>     {'rot_k': 3, 'flip_axis': None},
            >>>     {'rot_k': 0, 'flip_axis': (0,)},
            >>>     {'rot_k': 1, 'flip_axis': (0,)},
            >>>     {'rot_k': 2, 'flip_axis': (0,)},
            >>>     {'rot_k': 3, 'flip_axis': (0,)},
            >>>     # The rest of these dont result in any different data, but we need to test them
            >>>     {'rot_k': 0, 'flip_axis': (1,)},
            >>>     {'rot_k': 1, 'flip_axis': (1,)},
            >>>     {'rot_k': 2, 'flip_axis': (1,)},
            >>>     {'rot_k': 3, 'flip_axis': (1,)},
            >>>     {'rot_k': 0, 'flip_axis': (0, 1)},
            >>>     {'rot_k': 1, 'flip_axis': (0, 1)},
            >>>     {'rot_k': 2, 'flip_axis': (0, 1)},
            >>>     {'rot_k': 3, 'flip_axis': (0, 1)},
            >>> ]
            >>> results = []
            >>> for params in fliprot_params:
            >>>     tf = kwimage.Affine.fliprot(canvas_dsize=canvas_dsize, **params)
            >>>     annot2 = annot.warp(tf)
            >>>     annot3 = annot2.warp(tf.inv())
            >>>     #annot3 = inv_fliprot_annot(annot2, canvas_dsize=canvas_dsize, **params)
            >>>     results.append({
            >>>         'annot2': annot2,
            >>>         'annot3': annot3,
            >>>         'params': params,
            >>>         'tf': tf,
            >>>         'canvas_dsize': canvas_dsize,
            >>>     })
            >>> box = kwimage.Box.coerce([0, 0, W, H], format='xywh')
            >>> print('box = {}'.format(ub.urepr(box, nl=1)))
            >>> for result in results:
            >>>     params = result['params']
            >>>     warped = box.warp(result['tf'])
            >>>     print('---')
            >>>     print('params = {}'.format(ub.urepr(params, nl=1)))
            >>>     print('warped = {}'.format(ub.urepr(warped, nl=1)))
            >>>     print(ub.hzcat(['tf = ', ub.urepr(result['tf'], nl=1)]))

            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> S = max(W, H)
            >>> image1 = np.linspace(.1, .9, W * H).reshape((H, W))
            >>> image1 = kwimage.atleast_3channels(image1)
            >>> image1[0, :, 0] = 1
            >>> image1[:, 0, 2] = 1
            >>> image1[1, :, 1] = 1
            >>> image1[:, 1, 1] = 1
            >>> image1[3, :, 0] = 0.5
            >>> image1[:, 7, 1] = 0.5
            >>> pnum_ = kwplot.PlotNums(nCols=4, nSubplots=len(results))
            >>> # NOTE: setting new_dsize='positive' illustrates an issuew with
            >>> # the pixel interpretation.
            >>> new_dsize = (S, S)
            >>> #new_dsize = 'positive'
            >>> for result in results:
            >>>     image2 = kwimage.warp_affine(image1.copy(), result['tf'], dsize=new_dsize)
            >>>     image3 = kwimage.warp_affine(image2.copy(), result['tf'].inv(), dsize=new_dsize)
            >>>     annot2 = result['annot2']
            >>>     annot3 = result['annot3']
            >>>     #canvas1 = annot1.draw_on(image1.copy(), edgecolor='kitware_blue', fill=False)
            >>>     #canvas2 = annot2.draw_on(image2.copy(), edgecolor='kitware_green', fill=False)
            >>>     #canvas3 = annot3.draw_on(image3.copy(), edgecolor='kitware_red', fill=False)
            >>>     canvas = kwimage.stack_images([image1, image2, image3], axis=1, pad=1, bg_value='green')
            >>>     kwplot.imshow(canvas, pnum=pnum_(), title=ub.urepr(result['params'], nl=0, compact=1, nobr=1))
            >>> kwplot.show_if_requested()
        """
        import kwimage
        rot_k = rot_k % 4  # only 4 cases
        tf = None

        HALF_OFFSET = 1
        if HALF_OFFSET:
            half1 = kwimage.Affine.translate((.5, .5))
            tf = half1
        else:
            tf = kwimage.Affine.eye()

        if flip_axis is not None:
            canvas_w, canvas_h = canvas_dsize
            canvas_dims = (canvas_h, canvas_w)
            yx2 = [0, 0]

            # Make the flip matrix
            F = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
            for axis in flip_axis:
                mdim = 1 - axis
                F[mdim, mdim] *= -1
                # When an axis is flipped we have to translate it to adjust it
                # back to the first quadrent
                dim = axes[axis]
                yx2[dim] = canvas_dims[dim]

            x2, y2 = yx2[::-1]
            T2 = kwimage.Affine.translate((x2, y2))
            tf_flip = T2 @ F
            if tf is None:
                tf = tf_flip
            else:
                tf = tf_flip @ tf

        if rot_k != 0:
            # Construct the rotation
            # Should we add this as a rotate90 function that doesn't contain pi
            # approximations?
            # tau = np.pi * 2
            # theta = -(rot_k * tau / 4)
            # R = kwimage.Affine.rotate(theta=theta)
            if rot_k == 1:
                R = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])
            elif rot_k == 2:
                R = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])
            elif rot_k == 3:
                R = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]])
            else:
                raise AssertionError

            # The rotation will be about (0, 0), so to ensure the results stays
            # in the positive quadrent
            canvas_w, canvas_h = canvas_dsize
            if rot_k == 1:
                x2 = 0
                y2 = canvas_w
            elif rot_k == 2:
                x2 = canvas_w
                y2 = canvas_h
            elif rot_k == 3:
                x2 = canvas_h
                y2 = 0
            else:
                raise AssertionError
            T2 = kwimage.Affine.translate((x2, y2))

            # Rotate and translate the data back into the first quadrent
            tf_rot = T2 @ R

            if tf is None:
                tf = tf_rot
            else:
                tf = tf_rot @ tf

        if tf is None:
            tf = Affine.eye()
        else:
            if HALF_OFFSET:
                half2 = kwimage.Affine.translate((-.5, -.5))
                tf = half2 @ tf

        return tf


try:
    import sympy
    _RationalMatrixBase = sympy.Matrix
except Exception:
    sympy = None
    _RationalMatrixBase = object


class _RationalNDArray(_RationalMatrixBase):
    """
    Wraps sympy matrices to make it somewhat more compatible with numpy.

    Example:
        >>> # xdoctest: +REQUIRES(module:sympy)
        >>> from kwimage.transform import *  # NOQA
        >>> from kwimage.transform import _RationalNDArray
        >>> arr = np.random.rand(3, 3)
        >>> a = _RationalNDArray.from_numpy(arr)
        >>> b = np.random.rand(3, 3)
        >>> c = a @ b
        >>> c @ c.inv()
    """

    @classmethod
    def from_numpy(_RationalNDArray, arr):
        flat_rat = list(map(sympy.Rational, arr.ravel().tolist()))
        self = _RationalNDArray(flat_rat).reshape(*arr.shape)
        return self

    def __matmul__(self, other):
        if isinstance(other, np.ndarray):
            other = _RationalNDArray.from_numpy(other)
        return super().__matmul__(other)

    def __rmatmul__(self, other):
        if isinstance(other, np.ndarray):
            other = _RationalNDArray.from_numpy(other)
        return super().__matmul__(other)

    def numpy(self):
        return np.array(self.tolist()).astype(float)

    def ravel(self):
        return self.flat()

# Does not seem to be working out
# if 0:
#     try:
#         from mpmath import matrix as mp_matrix_base
#     except ImportError:
#         mp_matrix_base = object
#     class _mpmatrix(mp_matrix_base):
#         """
#         A compatability layer for mpmath matrix

#         Example:
#             >>> # xdoctest: +REQUIRES(module:mpmath)
#             >>> from kwimage.transform import _mpmatrix  # NOQA
#             >>> A = _mpmatrix(np.random.rand(3, 3))
#             >>> B = _mpmatrix(np.random.rand(3, 3))
#             >>> C = np.random.rand(3, 3)
#             >>> A @ B
#             >>> B @ A
#             >>> self = A
#             >>> other = C
#             >>> A.__matmul__(C)
#             >>> # C.__matmul__(A) not sure why this fails
#         """

#         @property
#         def shape(self):
#             return (self.rows, self.cols)

#         def __matmul__(self, other):
#             if isinstance(other, np.ndarray):
#                 other = _mpmatrix(other)
#             return _mpmatrix(mp_matrix_base.__matmul__(self, other))

#         def __rmatmul__(self, other):
#             if isinstance(other, np.ndarray):
#                 other = _mpmatrix(other)
#             return _mpmatrix(mp_matrix_base.__matmul__(other, self))

#         def numpy(self):
#             return np.array(self).reshape(*self.shape)


# def _ensure_iterablen(scalar, n):
#     try:
#         iter(scalar)
#     except TypeError:
#         return [scalar] * n
#     return scalar


def _ensure_iterable2(scalar):
    try:
        a, b = scalar
    except TypeError:
        a = b = scalar
    return a, b

if __name__ == '__main__':
    """
    CommandLine:
        python -m kwimage.transform all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class StackImagesCLI(scfg.DataConfig):
    """
    Stacks multiple images on disk into a single stacked image.
    """
    __command__ = 'stack_images'

    input_fpaths = scfg.Value(None, nargs='+', position=1, type=str, help=ub.paragraph(
        '''
        A list of input file paths, directories, or glob patterns. If a directory
        is specified, all files with a known image extension are included.
        This functionality requires `kwutil` to resolve glob patterns and directory inputs.
        '''))
    axis = scfg.Value('grid', help=ub.paragraph(
        '''
        The axis to stack over. Use `0` for vertical stacking, `1` for horizontal stacking,
        or `grid` to arrange images in a grid pattern. Default is `grid`.
        '''))
    pad = scfg.Value(None, help=ub.paragraph(
        '''
        The amount of padding (in pixels) to add between stacked images.
        If `None`, no padding is applied.
        '''
    ))
    out = scfg.Value(None, help=ub.paragraph(
        '''
        Path to save the output stacked image. If unspecified, a uses a
        hash-based filename (derived from the input image paths).
        '''
    ))

    @classmethod
    def main(StackImagesCLI, cmdline=1, **kwargs):
        """
        Example:
            >>> # xdoctest: +SKIP
            >>> # xdoctest: +REQUIRES(module:scriptconfig)
            >>> cmdline = 0
            >>> kwargs = dict(
            >>> )
            >>> main(cmdline=cmdline, **kwargs)
        """
        config = StackImagesCLI.cli(cmdline=cmdline, data=kwargs)
        import kwimage
        print('config = ' + ub.urepr(dict(config), nl=1))
        fpaths = config['input_fpaths']

        try:
            import kwutil
        except ImportError:
            import warnings
            warnings.warn('kwutil is not available; glob patterns and directory input may be limited.')
        else:
            # If available use kwutil to allow for a better
            from kwimage import im_io
            fpaths = kwutil.util_path.coerce_patterned_paths(fpaths, expected_extension=im_io.IMAGE_EXTENSIONS)

        images = [kwimage.imread(p) for p in ub.ProgIter(fpaths, desc='read images')]

        if config['axis'] == 'grid':
            canvas = kwimage.stack_images_grid(images, pad=config['pad'])
        else:
            canvas = kwimage.stack_images(images, axis=config['axis'], pad=config['pad'])

        out_fpath = config['out']
        if out_fpath is None:
            out_fpath = 'stack_' + ub.hash_data(fpaths)[0:16] + '.png'
        print(f'write to: {out_fpath}')
        kwimage.imwrite(out_fpath, canvas)


__cli__ = StackImagesCLI

if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/kwimage/kwimage/cli/stack_images.py
        python -m stack_images
    """
    __cli__.main()



from typing import Iterable
from numpy import ndarray
from typing import Tuple
from numbers import Number
from typing import List
import ubelt as ub
from _typeshed import Incomplete

__todo__: str
BASE_COLORS: Incomplete
TABLEAU_COLORS: Incomplete
XKCD_COLORS: Incomplete
CSS4_COLORS: Incomplete
KITWARE_COLORS: Incomplete


class Color(ub.NiceRepr):
    color01: Incomplete
    space: str

    def __init__(self,
                 color: Color | Iterable[int | float] | str,
                 alpha: float | None = None,
                 space: str | None = None,
                 coerce: bool = True) -> None:
        ...

    @classmethod
    def coerce(cls, data, **kwargs):
        ...

    def __nice__(self):
        ...

    def forimage(self,
                 image: ndarray,
                 space: str = 'auto') -> Tuple[Number, ...]:
        ...

    def ashex(self, space: None | str = None) -> str:
        ...

    def as255(
        self,
        space: None | str = None
    ) -> Tuple[int, int, int] | Tuple[int, int, int, int]:
        ...

    def as01(
        self,
        space: None | str = None
    ) -> Tuple[float, float, float] | Tuple[float, float, float, float]:
        ...

    @classmethod
    def named_colors(cls) -> List[str]:
        ...

    @classmethod
    def distinct(Color,
                 num,
                 existing: Incomplete | None = ...,
                 space: str = ...,
                 legacy: str = ...,
                 exclude_black: bool = ...,
                 exclude_white: bool = ...) -> List[Tuple]:
        ...

    @classmethod
    def random(Color,
               pool: str = ...,
               with_alpha: int = ...,
               rng: Incomplete | None = ...) -> Color:
        ...

    def distance(self, other: Color, space: str = 'lab') -> float:
        ...

    def interpolate(self,
                    other: Color,
                    alpha: float | List[float] = 0.5,
                    ispace: str | None = None,
                    ospace: str | None = None) -> Color | List[Color]:
        ...

    def to_image(self, dsize: Tuple[int, int] = ...):
        ...

    def adjust(self, saturate: float = 0, lighten: float = 0):
        ...



import ubelt as ub
import sys
import kwarray
import numbers
import numpy as np
# from collections import abc
# import abc

# try:
#     import torch
# except Exception:
#     torch = None
#     ARRAY_TYPES = (np.ndarray,)
# else:
#     ARRAY_TYPES = (np.ndarray, torch.Tensor)


def isinstance_arraytypes(obj):
    """
    workaround so we dont need to import torch at the global level
    """
    torch = sys.modules.get('torch', None)
    if torch is None:
        return isinstance(obj, np.ndarray)
    else:
        return isinstance(obj, (np.ndarray, torch.Tensor))


# class Spatial(ub.NiceRepr, abc.ABC):
class Spatial(ub.NiceRepr):
    """
    Abstract base class defining the spatial annotation API
    """

    # @abc.abstractmethod
    # def translate(self, offset, output_dims=None):
    #     raise NotImplementedError

    # @abc.abstractmethod
    # def scale(self, factor, output_dims=None):
    #     raise NotImplementedError

    # @abc.abstractmethod
    # def warp(self, transform, input_dims=None, output_dims=None, inplace=False):
    #     raise NotImplementedError

    # def draw(self, image=None, **kwargs):
    #     # If draw doesnt exist use draw_on
    #     import numpy as np
    #     if image is None:
    #         dims = self.bounds
    #         shape = tuple(dims) + (4,)
    #         image = np.zeros(shape, dtype=np.float32)
    #     image = self.draw_on(image, **kwargs)
    #     import kwplot
    #     kwplot.imshow(image)

    # @abc.abstractmethod
    # def draw_on(self, image):
    #     """
    #     TODO -
    #         - [ ] Choose 1:
    #             Should accept either uint255 or float01, and should return the
    #             same kind
    #     """
    #     raise NotImplementedError

    # @abc.abstractmethod
    # def tensor(self, device=ub.NoParam):
    #     raise NotImplementedError

    # @abc.abstractmethod
    # def numpy(self):
    #     raise NotImplementedError

    # @classmethod
    # @abc.abstractmethod
    # def random(cls):
    #     raise NotImplementedError


# class ListProxy(abc.MutableSequence):
class _ExperimentalListProxy:
    """
    We may modify this implementation to be more generic in the future and
    directly inherit from :class:`abc.MutableSequence`. This intermediate form
    helps preserve backwards compatibility of ObjectList, but it is unclear if
    that is neeeded.

    Partially generated via ChatGPT.

    Requires that the inheriting class has a ``data`` attribute.
    """

    def __getitem__(self, index):
        """Retrieve an item by its index."""
        return self.data[index]

    def __setitem__(self, index, value):
        """Update an item at the specified index."""
        self.data[index] = value

    def __delitem__(self, index):
        """Delete an item at the specified index."""
        del self.data[index]

    def __len__(self):
        """Return the length of the sequence."""
        return len(self.data)

    def insert(self, index, value):
        """Insert an item at a specific index."""
        self.data.insert(index, value)

    def append(self, value):
        """Add an item to the end of the sequence."""
        self.data.append(value)

    def clear(self):
        """Remove all items from the sequence."""
        self.data.clear()

    def reverse(self):
        """Reverse the sequence in place."""
        self.data.reverse()

    def extend(self, other):
        """Extend the sequence by appending elements from an iterable."""
        self.data.extend(other)

    def pop(self, index=-1):
        """Remove and return an item at the given index."""
        return self.data.pop(index)

    def remove(self, value):
        """Remove the first occurrence of a value."""
        self.data.remove(value)

    def __iadd__(self, other):
        """Support in-place addition (+=) to extend the sequence."""
        return self.data.__iadd__(other)

    def __contains__(self, item):
        """Check if the item is in the sequence."""
        return item in self.data

    def __iter__(self):
        """Return an iterator over the sequence."""
        return iter(self.data)

    def __reversed__(self):
        """Return a reverse iterator over the sequence."""
        return self.data.__reversed__()

    def index(self, value, start=0, stop=None):
        """
        Return the index of the first occurrence of a value.
        Raise ValueError if the value is not present.
        """
        return self.data.index(value, start, stop)

    def count(self, value):
        """Return the number of occurrences of a value."""
        return self.data.count(value)


class ObjectList(Spatial, _ExperimentalListProxy):
    """
    Stores a list of potentially heterogenous structures, each item usually
    corresponds to a different object.

    Example:
        >>> # TODO: better test that flexes the Spatial inheritence and
        >>> # method beyond those in _ExperimentalListProxy
        >>> from kwimage.structs._generic import *  # NOQA
        >>> class SpatialList(ObjectList):
        >>>    ...
        >>> self = SpatialList([])
        >>> self.append(1)
        >>> self.extend([1])
        >>> 1 in self
    """

    # __slots__ = ('data', 'meta',)

    def __init__(self, data, meta=None):
        if meta is None:
            meta = {}
        self.data = data
        self.meta = meta

    def __len__(self):
        return len(self.data)

    @property
    def shape(self):
        return (len(self),)

    @property
    def dtype(self):
        try:
            return self.data.dtype
        except Exception:
            print('kwimage._generic: no dtype for ' + str(type(self.data)))
            raise

    def __nice__(self):
        return 'n={}'.format(len(self))

    def translate(self, offset, output_dims=None, inplace=False):
        newdata = [None if item is None else
                   item.translate(offset, output_dims=output_dims,
                                  inplace=inplace)
                   for item in self.data]
        return self.__class__(newdata, self.meta)

    def scale(self, factor, output_dims=None, inplace=False):
        newdata = [None if item is None else
                   item.scale(factor, output_dims=output_dims, inplace=inplace)
                   for item in self.data]
        return self.__class__(newdata, self.meta)

    def warp(self, transform, input_dims=None, output_dims=None, inplace=False):
        if inplace:
            for item in self.data:
                if item is not None:
                    item.warp(transform, input_dims=input_dims,
                              output_dims=output_dims, inplace=inplace)
            return self
        else:
            newdata = [None if item is None else
                       item.warp(transform, input_dims=input_dims,
                                 output_dims=output_dims, inplace=inplace)
                       for item in self.data]
            return self.__class__(newdata, self.meta)

    def apply(self, func):
        newdata = [None if item is None else func(item) for item in self.data]
        return self.__class__(newdata, self.meta)

    def to_coco(self, style='orig'):
        for item in self.data:
            if item is None:
                yield None
            else:
                yield item.to_coco(style=style)

    def compress(self, flags, axis=0):
        assert axis == 0
        newdata = list(ub.compress(self.data, flags))
        return self.__class__(newdata, self.meta)

    def take(self, indices, axis=0):
        assert axis == 0
        newdata = list(ub.take(self.data, indices))
        return self.__class__(newdata, self.meta)

    def draw(self, **kwargs):
        """
        Generic draw method for list of spatial annotations
        """
        if 'setlim' in kwargs:
            import warnings
            warnings.warn('ObjectList currently need special handling for setlim kwarg')
        patches = []
        for item in self.data:
            if item is not None:
                patch = item.draw(**kwargs)
                patches.append(patch)
        return patches

    def draw_on(self, image, **kwargs):
        """
        TODO:
            document fastdraw - it flattens all subobjects into the same layer
            and then does any alpha blending. Is there a better name?
        """
        import kwimage

        # Take care of data prep before looping
        alpha = kwargs.get('alpha', None)
        copy = kwargs.pop('copy', False)

        has_float_alpha = alpha is not None and alpha < 1.0

        if has_float_alpha:
            image = kwimage.ensure_float01(image, copy=copy)
        elif copy:
            image = image.copy()

        # Check if we want to do the fastdraw hack
        fastdraw = kwargs.pop('fastdraw', 'auto')
        if fastdraw == 'auto':
            fastdraw = has_float_alpha

        if fastdraw:
            # Fast draw hack will remove the alpha from the subcalls to
            # draw_on, instead we will draw with full alpha on an empty canvas
            # and then blend together everything at the end.
            orig_canvas = image
            overlay_canvas = np.zeros_like(image, shape=(image.shape[0:2] + (4,)))
            image = overlay_canvas
            kwargs['alpha'] = None

        # Handle per-instance arguments
        # If color is given an it corresponds to each subitem
        # then pass the appropriate arg to each subitem
        instkw_list = [{} for _ in range(len(self.data))]
        selflen = len(self.data)
        _handle_perinstance_color_arg(selflen, instkw_list, kwargs, 'color')
        _handle_perinstance_color_arg(selflen, instkw_list, kwargs, 'edgecolor')
        _handle_perinstance_color_arg(selflen, instkw_list, kwargs, 'facecolor')

        for item, instkw in zip(self.data, instkw_list):
            if item is not None:
                image = item.draw_on(image=image, **kwargs, **instkw)

        if fastdraw:
            # Blend the empty canvas back onto
            overlay_canvas = image
            overlay_canvas[..., 3] *= alpha
            image = kwimage.overlay_alpha_images(overlay_canvas, orig_canvas)
            pass

        return image

    def tensor(self, device=ub.NoParam):
        return self.apply(lambda item: item.tensor(device))

    def numpy(self):
        return self.apply(lambda item: item.numpy())

    @classmethod
    def concatenate(cls, items, axis=0):
        """
        Args:
            items (Sequence[ObjectList]): multiple object lists of the same type
            axis (int | None): unused, always implied to be axis 0

        Returns:
            ObjectList: combined object list

        Example:
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> import kwimage
            >>> cls = kwimage.MaskList
            >>> sub_cls = kwimage.Mask
            >>> item1 = cls([sub_cls.random(), sub_cls.random()])
            >>> item2 = cls([sub_cls.random()])
            >>> items = [item1, item2]
            >>> new = cls.concatenate(items)
            >>> assert len(new) == 3
        """
        if len(items) == 0:
            new = cls([])
        else:
            newdata = []
            for item in items:
                newdata.extend(item.data)

        newmeta = items[0].meta
        new = cls(newdata, newmeta)
        return new

    def is_tensor(cls):
        raise NotImplementedError

    def is_numpy(cls):
        raise NotImplementedError

    # @classmethod
    # def random(cls):
    #     raise NotImplementedError


def _handle_perinstance_color_arg(selflen, instkw_list, kwargs, argname):
    """
    helper to expand any color argument into multiple color arguments for each
    instance handled by the generic draw on method. This allows the user to
    specify a list of colors for each member, or a single color to be applied
    to everyone.
    """
    if argname in kwargs:
        color = kwargs.pop(argname, None)
        if (ub.iterable(color) and len(color) == selflen and
             len(color) > 0 and not isinstance(ub.peek(color), numbers.Number)):
            for d, c in zip(instkw_list, color):
                d[argname] = c
        else:
            kwargs[argname] = color


def _coerce_color_list_for(image, color, num):
    """
    Get a list of colors for each item
    """
    import kwimage
    if (ub.iterable(color) and len(color) == num and
         len(color) > 0 and not isinstance(ub.peek(color), numbers.Number)):
        color_list = [kwimage.Color(c).forimage(image) for c in color]
    else:
        color_list = [kwimage.Color(color).forimage(image)] * num
    return color_list


def _consistent_dtype_fixer(data):
    """
    helper for ensuring out.dtype == in.dtype
    """
    import kwimage
    if data.dtype.kind == 'f':
        return kwimage.ensure_float01
    elif data.dtype.kind == 'u':
        return kwimage.ensure_uint255
    else:
        raise TypeError(data.dtype)


def _safe_take(data, indices, axis):
    if data is None:
        return data
    try:
        return data.take(indices, axis=axis)
    except (TypeError, AttributeError):
        return kwarray.ArrayAPI.take(data, indices, axis=axis)


def _safe_compress(data, flags, axis):
    if data is None:
        return data
    try:
        return data.compress(flags, axis=axis)
    except (TypeError, AttributeError):
        return kwarray.ArrayAPI.compress(data, flags, axis=axis)


def _issubclass2(child, parent):
    """
    Uses string comparisons to avoid ipython reload errors.
    Much less robust though.
    """
    # String comparison
    if child.__name__ == parent.__name__:
        if child.__module__ == parent.__module__:
            return True
    # Recurse through classes of obj
    return any(_issubclass2(base, parent) for base in child.__bases__)


def _isinstance2(obj, cls):
    """
    Uses string comparisons to avoid ipython reload errors.
    Much less robust though.

    Example:
        import kwimage
        from kwimage.structs import _generic
        cls = kwimage.structs._generic.ObjectList
        obj = kwimage.MaskList([])
        _generic._isinstance2(obj, cls)

        _generic._isinstance2(kwimage.MaskList([]), _generic.ObjectList)

        dets = kwimage.Detections(
            boxes=kwimage.Boxes.random(3).numpy(),
            class_idxs=[0, 1, 1],
            segmentations=kwimage.MaskList([None] * 3)
        )
    """
    if isinstance(obj, cls):
        return True
    try:
        return _issubclass2(obj.__class__, cls)
    except Exception:
        return False
    return False


def _setlim(xmin, ymin, xmax, ymax, setlim=False, ax=None):
    """
    Helper for setlim argument for draw function.
    TODO: replace with _setlim2 and move to kwplot
    """
    if ax is None:
        from matplotlib import pyplot as plt
        ax = plt.gca()

    if isinstance(setlim, str):
        if setlim == 'grow':
            # only allow growth
            x1_, x2_ = ax.get_xlim()
            xmin = min(x1_, xmin)
            xmax = max(x2_, xmax)

            y1_, y2_ = ax.get_ylim()
            ymin = min(y1_, ymin)
            ymax = max(y2_, ymax)
        else:
            raise KeyError(setlim)

    if isinstance(setlim, float):
        w = abs(xmax - xmin)
        h = abs(ymax - ymin)
        xpad = ((w * setlim) - w) / 2
        ypad = ((h * setlim) - h) / 2

        xmin = xmin - xpad
        ymin = ymin - ypad

        xmax = xmax + xpad
        ymax = ymax + ypad

    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)


def _setlim2(xmin=None, ymin=None, xmax=None, ymax=None, setlim=False, ax=None):
    """
    Helper for setlim argument for draw function.
    TODO: move to kwplot
    """
    if ax is None:
        from matplotlib import pyplot as plt
        ax = plt.gca()

    # Parse options
    growonly_mode = False
    scale_factor = None
    if isinstance(setlim, str):
        if setlim == 'grow':
            growonly_mode = True
        else:
            raise KeyError(setlim)
    elif isinstance(setlim, float):
        scale_factor = setlim

    # The logic is the same for both x and y so abstract it a bit.
    requests = [
        {'request_min': xmin, 'request_max': xmax, 'getter': ax.get_xlim, 'setter': ax.set_xlim},
        {'request_min': ymin, 'request_max': ymax, 'getter': ax.get_ylim, 'setter': ax.set_ylim},
    ]

    for request in requests:
        _setlim_for_dim(**request, scale_factor=scale_factor, growonly_mode=growonly_mode)


def _setlim_for_dim(request_min, request_max, getter, setter, scale_factor, growonly_mode):
    """
    Example:
        >>> def getter():
        >>>     return (20, 99)
        >>> def setter(new_min, new_max):
        >>>     print('new min/max', new_min, new_max)
        >>> # Todo: run over a grid
        >>> request_min = 20
        >>> request_max = 232
        >>> scale_factor = 1.2
        >>> growonly_mode = 0
        >>> _setlim_for_dim(request_min, request_max, getter, setter, scale_factor, growonly_mode)
    """
    # If the min/max value is not given we "pin" it to prevent it from
    # changing.
    pin_min = request_min is None
    pin_max = request_max is None

    if pin_min and pin_max:
        # optimization, dont need to do anything if there is no request
        return

    orig_min, orig_max = getter()
    if pin_min:
        request_min = orig_min
    if pin_max:
        request_max = orig_max

    if growonly_mode:
        if not pin_min:
            request_min = min(orig_min, request_min)
        if not pin_max:
            request_max = max(orig_max, request_max)

    if scale_factor is not None:
        extent = abs(request_max - request_min)
        pad = ((extent * scale_factor) - extent) / 2
        if not pin_min:
            request_min = request_min - pad
        if not pin_max:
            request_max = request_max + pad

    setter(request_min, request_max)


def _handle_color_args_for(color, alpha, border, fill, edgecolor, facecolor, image):
    """
    Common logic for boxes and polygons
    """
    import kwimage
    if facecolor is None:
        if fill:
            facecolor = color
    elif facecolor is True:
        facecolor = color
    else:
        facecolor = kwimage.Color(facecolor, alpha=alpha).forimage(image)

    if edgecolor is None:
        if border:
            edgecolor = color
    elif edgecolor is True:
        edgecolor = color
    else:
        edgecolor = kwimage.Color(edgecolor, alpha=alpha).forimage(image)



from numpy import ndarray
import skimage.transform
from typing import Dict
from typing import Tuple
import ubelt as ub
from _typeshed import Incomplete

import affine

__docstubs__: str


class Transform(ub.NiceRepr):
    ...


class Matrix(Transform):
    matrix: Incomplete

    def __init__(self, matrix) -> None:
        ...

    def __nice__(self):
        ...

    @property
    def shape(self):
        ...

    def __json__(self):
        ...

    @classmethod
    def coerce(cls, data: Incomplete | None = ..., **kwargs):
        ...

    def __array__(self):
        ...

    def __imatmul__(self, other) -> None:
        ...

    def __matmul__(self, other):
        ...

    def is_rational(self):
        ...

    def inv(self) -> Matrix:
        ...

    @property
    def T(self):
        ...

    def det(self) -> float:
        ...

    @classmethod
    def eye(cls, shape: Incomplete | None = ..., rng: Incomplete | None = ...):
        ...

    @classmethod
    def random(cls,
               shape: Incomplete | None = ...,
               rng: Incomplete | None = ...):
        ...

    def __getitem__(self, index):
        ...

    def rationalize(self):
        ...

    def astype(self, dtype: type):
        ...

    def isclose_identity(self, rtol: float = ..., atol: float = ...):
        ...


class Linear(Matrix):
    ...


class Projective(Linear):

    @classmethod
    def fit(cls, pts1: ndarray, pts2: ndarray) -> Projective:
        ...

    @classmethod
    def projective(cls,
                   scale: Incomplete | None = ...,
                   offset: Incomplete | None = ...,
                   shearx: Incomplete | None = ...,
                   theta: Incomplete | None = ...,
                   uv: Incomplete | None = ...,
                   about: Incomplete | None = ...):
        ...

    @classmethod
    def coerce(cls, data: Incomplete | None = ..., **kwargs) -> Projective:
        ...

    def is_affine(self) -> bool:
        ...

    def to_skimage(self) -> skimage.transform.AffineTransform:
        ...

    @classmethod
    def random(cls,
               shape: Incomplete | None = ...,
               rng: Incomplete | None = ...,
               **kw):
        ...

    def decompose(self) -> Dict:
        ...


class Affine(Projective):

    @property
    def shape(self):
        ...

    def __json__(self):
        ...

    def concise(self) -> Dict:
        ...

    @classmethod
    def from_shapely(cls, sh_aff):
        ...

    @classmethod
    def from_affine(cls, aff):
        ...

    @classmethod
    def from_gdal(cls, gdal_aff):
        ...

    @classmethod
    def from_skimage(cls, sk_aff):
        ...

    @classmethod
    def coerce(cls, data: Incomplete | None = ..., **kwargs) -> Affine:
        ...

    def eccentricity(self) -> float:
        ...

    def to_affine(self) -> affine.Affine:
        ...

    def to_gdal(self) -> Tuple[float, float, float, float, float, float]:
        ...

    def to_shapely(self) -> Tuple[float, float, float, float, float, float]:
        ...

    def to_skimage(self) -> skimage.transform.AffineTransform:
        ...

    @classmethod
    def scale(cls, scale: float | Tuple[float, float]) -> Affine:
        ...

    @classmethod
    def translate(cls, offset: float | Tuple[float, float]) -> Affine:
        ...

    @classmethod
    def rotate(cls, theta: float) -> Affine:
        ...

    @classmethod
    def random(cls,
               shape: Incomplete | None = ...,
               rng: Incomplete | None = ...,
               **kw) -> Affine:
        ...

    @classmethod
    def random_params(cls, rng: Incomplete | None = ..., **kw) -> Dict:
        ...

    def decompose(self) -> Dict:
        ...

    @classmethod
    def affine(cls,
               scale: float | Tuple[float, float] | None = None,
               offset: float | Tuple[float, float] | None = None,
               theta: float | None = None,
               shear: float | None = None,
               about: float | Tuple[float, float] | None = None,
               shearx: float | None = None,
               array_cls: Incomplete | None = ...,
               math_mod: Incomplete | None = ...,
               **kwargs) -> Affine:
        ...

    @classmethod
    def fit(cls, pts1: ndarray, pts2: ndarray) -> Affine:
        ...



import kwimage
import kwarray
import numpy as np
import ubelt as ub
import pytest


def image_variations(image_basis):
    """
    Helper to make several variations of image inputs for opencv with different
    dtypes etc..
    """
    try:
        import cv2  # NOQA
    except ImportError:
        pytest.skip('requires cv2')
    rng = kwarray.ensure_rng(0)

    if image_basis is None:
        image_basis = {
            'dims': [(32, 32), (37, 41), (53, 31)],
            'channels': [None, 1, 3, 4, 20, 1024],
            'dtype': ['uint8', 'int64', 'float32', 'float64'],
        }

    # TODO: how to specify conditionals?
    # conditionals = {
    #     np.uint8
    # }

    for imgkw in list(ub.named_product(image_basis)):
        if imgkw['channels'] is None:
            shape = imgkw['dims']
        else:
            shape = imgkw['dims'] + (imgkw['channels'],)
        dtype = np.dtype(imgkw['dtype'])
        img = rng.rand(*shape)
        if dtype.kind in {'i', 'u'}:
            img = img * 255
        img = img.astype(dtype)
        yield imgkw, img


def test_imresize_multi_channel():
    """
    Test that imresize works with multiple channels in various configurations
    """
    try:
        import cv2  # NOQA
    except ImportError:
        pytest.skip('requires cv2')

    resize_kw_basis = {
        'dsize': [
            # (10, 10),
            (60, 60)
        ],
        'interpolation': [
            'auto',
            # 'area',
            # 'linear',
            # 'cubic',
            'nearest'
        ],
        'antialias': [
            True,
            # False,
        ]
    }

    image_basis = {
        'dims': [
            # (32, 32),
            # (37, 41),
            (53, 31)
        ],
        'channels': [
            None,
            1,
            3,
            4,
            20,
            1024,
        ],
        'dtype': [
            'uint8',
            'float32',
            'float64'
        ],
    }

    resize_kw_list = list(ub.named_product(resize_kw_basis))

    failures = []
    success = []
    import timerit
    ti = timerit.Timerit(1, bestof=1, verbose=0)

    for imgkw, img in image_variations(image_basis):
        for resize_kw in resize_kw_list:
            params = dict(resize_kw=resize_kw, imgkw=imgkw)
            try:
                label = ub.urepr(params, nl=0, nobr=True, si=1, sv=1, kvsep='=', itemsep='')
                for timer in ti.reset(label):
                    with timer:
                        kwimage.imresize(img, **resize_kw)
            except Exception:
                failures.append(label)
                print('FAILED = {!r}'.format(label))
                raise
            else:
                success.append(label)

    print('n_pass = {}'.format(len(success)))
    print('n_fail = {}'.format(len(failures)))
    print('failures = {}'.format(ub.urepr(failures, nl=1)))


def test_resize_shapes():
    try:
        import cv2  # NOQA
    except ImportError:
        pytest.skip('requires cv2')
    import kwimage
    data = np.random.rand(32, 32)
    new = kwimage.imresize(data, scale=2)
    assert new.shape == (64, 64)

    data = np.random.rand(32, 32, 1)
    new = kwimage.imresize(data, scale=2)
    assert new.shape == (64, 64, 1)

    data = np.random.rand(32, 32, 3)
    new = kwimage.imresize(data, scale=2)
    assert new.shape == (64, 64, 3)


def test_resize_bool():
    try:
        import cv2  # NOQA
    except ImportError:
        pytest.skip('requires cv2')
    import kwimage
    img = (np.random.rand(32, 32, 1) > 0.5).astype(bool)
    new = kwimage.imresize(img, scale=2)
    assert new.dtype.kind == 'b'



def test_geotiff():
    """
    """
    from kwimage import im_io
    if not im_io._have_gdal():
        import pytest
        pytest.skip()

    import kwimage
    import ubelt as ub
    dpath = ub.Path.appdir('kwimage/tests/geotiff').ensuredir()
    fpath = dpath / 'dummy_geotiff.tif'

    # Make a random polygon in world space
    lat_max = 180 - 1e-7
    lon_max = 90 - 1e-7
    wld_poly = kwimage.Polygon.random().translate((-.5, -.5)).scale(2).scale((lat_max, lon_max))

    # With bounds that are associated to an image of size
    img_dsize = (640, 480)
    # Make a dummy geotiff
    _imdata = kwimage.grab_test_image("amazon")
    imdata = kwimage.imresize(_imdata, dsize=img_dsize)
    imdata = kwimage.ensure_float01(imdata)

    # Given this information, construct the geotiff metadata
    wld_box = wld_poly.box()
    img_box = kwimage.Box.from_dsize(img_dsize)
    wld_corners = wld_box.corners()
    img_corners = img_box.corners().astype(float)

    # The geotransform
    tf_wld_from_img = kwimage.Affine.fit(img_corners, wld_corners)

    # The CRS should be CRS-84
    from osgeo import osr
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(4326)
    srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    crs = srs.ExportToWkt()

    nodata_value = 0
    imdata = kwimage.ensure_uint255(imdata)

    # Add in some nodata_value
    imdata[-10:, 10:] = nodata_value
    imdata[0:10:, -200:-180] = nodata_value
    imdata = imdata[:, :, 0:3]

    metadata = {
        'SENSOR': 'kwimage-demo',
    }

    kwimage.imwrite(
        fpath,
        imdata,
        backend="gdal",
        nodata_value=nodata_value,
        crs=crs,
        transform=tf_wld_from_img,
        overviews="auto",
        metadata=metadata,
    )

    from osgeo import gdal
    import os
    dset = gdal.Open(os.fspath(fpath), gdal.GA_ReadOnly)
    info = gdal.Info(dset, format='json', allMetadata=True, listMDD=True)
    print('info = {}'.format(ub.urepr(info, nl=True)))



"""
Helpers that may be required across different backends.
"""


def _coerce_warp_dsize_inputs(dsize, input_dsize, transform, require_warped_info=False):
    """
    Given a warp operation, we will often need to preallocate size for the
    destination canvas. This may be specified by the user, but it is helpful to
    be able to precalculate it based on the transform and a specified
    heuristic.

    Ignore:
        import kwimage
        from kwimage._common import *  # NOQA
        from kwimage._common import _coerce_warp_dsize_inputs
        input_dsize = (128, 128)
        dsize = (128, 128)
        transform = kwimage.Affine.coerce({'scale': 0.2})

        basis = {
            'input_dsize': [
                (128, 128),
            ],
            'dsize': [
                (128, 128),
                'auto',
                'content',
            ],
            'transform': list(map(kwimage.Affine.coerce, [
                {'scale': 0.5},
                {'scale': 2.0},
                {'scale': 2.0, 'offset': -100},
            ])),
        }

        import rich
        for kwargs in ub.named_product(**basis):
            info = _coerce_warp_dsize_inputs(**kwargs)
            row = {'params': kwargs, 'result': info}
            rich.print(f'row = {ub.urepr(row, nl=2)}')
            ...
    """
    import numpy as np
    import kwimage

    w, h = input_dsize
    dsize

    # FIXME: we may need to modify this based on the origin convention for
    # correctness.

    if isinstance(dsize, str) or require_warped_info:
        # calculate dimensions needed for auto/max/try_large_warp
        input_box = kwimage.Boxes(np.array([[0, 0, w, h]]), 'xywh')
        warped_box = input_box.warp(transform)
        max_dsize = tuple(map(int, warped_box.to_xywh().quantize().data[0, 2:4]))
        new_origin = warped_box.to_ltrb().data[0, 0:2]
        if 0:
            # import rich
            # print('---------')
            # rich.print(f'box={box}')
            # rich.print(transform)
            # rich.print(f'warped_box={warped_box}')
            # TODO: should we enable this?  This seems to break if there is an
            # axis or orientation flip because the Boxes was designed with
            # slices in mind, so we need to maintain orientation information to
            # handle this correctly.
            warped_box._ensure_nonnegative_extent(inplace=True)
            # rich.print(f'warped_box={warped_box}')
            warped_box = warped_box.to_ltrb()
            # rich.print(f'warped_box={warped_box}')
            warped_box = warped_box.to_xywh().quantize()
            # rich.print(f'warped_box={warped_box}')
            max_dsize = tuple(map(int, warped_box.data[0, 2:4]))
            # print('warped_box = {}'.format(ub.urepr(warped_box, nl=1)))
            # print('max_dsize = {}'.format(ub.urepr(max_dsize, nl=1)))
    else:
        max_dsize = None
        new_origin = None

    transform_ = transform

    if dsize is None:
        # If unspecified, leave the canvas size unchanged
        dsize = (w, h)
    elif isinstance(dsize, str):
        # Handle special "auto-compute" dsize keys
        if dsize in {'positive', 'auto'}:
            quantized_warped_box = warped_box.to_ltrb().quantize()
            dsize = tuple(map(int, quantized_warped_box.data[0, 2:4]))
            if 0:
                affine_params = None
                # rich.print('affine_params = {}'.format(ub.urepr(affine_params, nl=1)))
                if affine_params is None:
                    affine_params = transform_.decompose()
                sx, sy = affine_params['scale']
                new_w, new_h = dsize
                if sx < 0:
                    new_w += 1
                if sy < 0:
                    new_h += 1
                dsize = (new_w, new_h)
        elif dsize in {'content', 'max'}:
            dsize = max_dsize
            transform_ = kwimage.Affine.translate(-new_origin) @ transform
            new_origin = np.array([0, 0])
        else:
            raise KeyError('Unknown dsize={}'.format(dsize))

    dsize_inputs = {
        'max_dsize': max_dsize,
        'new_origin': new_origin,
        'transform': transform_,
        'dsize': dsize,
    }
    return dsize_inputs



"""
mkinit ~/code/kwimage/kwimage/algo/__init__.py -w --relative
"""
from .algo_nms import (available_nms_impls, daq_spatial_nms,
                       non_max_supression,)

__all__ = ['available_nms_impls', 'daq_spatial_nms', 'non_max_supression']



# Helpers for typing

from skimage.transform import _geometric

try:
    SKImageGeometricTransform = _geometric._GeometricTransform
except AttributeError:
    # Older version compatability
    SKImageGeometricTransform = _geometric.GeometricTransform


__all__ = [
    'SKImageGeometricTransform'
]





def test_mask_with_bool_data():
    """
    Ensure that `to_multi_polygon` doesn't break when the mask is a boolean
    type. We mainly just run these to ensure there is no crash.
    """
    from kwimage._backend_info import _have_cv2
    if not _have_cv2():
        import pytest
        pytest.skip('requires cv2')
    import kwimage
    import ubelt as ub
    import numpy as np
    import kwarray
    rng = kwarray.ensure_rng(42134)
    mask_data = rng.rand(32, 32) > 0.5
    mask = kwimage.Mask(mask_data, 'c_mask')

    multi_poly = mask.to_multi_polygon()
    hull = mask.get_convex_hull()
    assert len(multi_poly) > 0
    assert isinstance(hull, np.ndarray)

    if ub.modname_to_modpath('kwimage_ext'):
        try:
            assert mask.iou(mask) > 0.999999
        except NotImplementedError:
            # kwimage_ext is likely not built correctly for this platform in
            # this case.
            ...

    coco_mask = mask.to_coco()
    assert isinstance(coco_mask, dict)
    assert mask.shape == (32, 32)
    assert mask.area > 0

    mask.translate(3)
    if ub.modname_to_modpath('torch'):
        import torch  # NOQA
        mask.warp(kwimage.Affine.eye(), output_dims='same')
        mask.scale(2.3)
    mask.get_patch()



def test_demodata_headers():
    """
    Check that the extensions on all of our demo images are accurate
    """
    from PIL import Image
    from os.path import splitext
    import kwimage
    import ubelt as ub

    if not ub.argflag('--network'):
        import pytest
        pytest.skip('requires network')

    for key in kwimage.grab_test_image.keys():
        fpath = kwimage.grab_test_image_fpath(key)
        img = Image.open(fpath)
        try:
            header_format = img.format
            ext_format = splitext(fpath)[-1].upper()[1:]

            if ext_format == 'JPG':
                ext_format = 'JPEG'

            if ext_format != header_format:
                print(img.format)
                print('fpath = {!r}'.format(fpath))
                print('--')

            assert ext_format == header_format, 'format should agree with ext'
        finally:
            img.close()



import scriptconfig as scfg
from _typeshed import Incomplete


class StackImagesConfig(scfg.DataConfig):
    input_fpaths: Incomplete
    axis: Incomplete
    pad: Incomplete
    out: Incomplete


def main(cmdline: int = ..., **kwargs) -> None:
    ...
